#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "23/11/2020"


from edna2.tasks.AbstractTask import AbstractTask

# aimedCompleteness : XSDataDouble optional
# aimedIOverSigma : XSDataDouble optional
# aimedRedundancy : XSDataDouble optional
# aimedResolution : XSDataDouble optional
# anomalousData : XSDataBoolean optional
# beamExposureTime : XSDataTime
# beamMaxExposureTime : XSDataTime optional
# beamMinExposureTime : XSDataTime optional
# bestFileContentDat : XSDataString
# bestFileContentHKL : XSDataString []
# bestFileContentPar : XSDataString
# complexity : XSDataString optional
# crystalAbsorbedDoseRate : XSDataAbsorbedDoseRate optional
# crystalShape : XSDataDouble optional
# crystalSusceptibility : XSDataDouble optional
# detectorDistanceMax : XSDataLength optional
# detectorDistanceMin : XSDataLength optional
# detectorType : XSDataString
# doseLimit : XSDataDouble optional
# goniostatMaxRotationSpeed : XSDataAngularSpeed optional
# goniostatMinRotationWidth : XSDataAngle optional
# minTransmission : XSDataDouble optional
# numberOfCrystalPositions : XSDataInteger optional
# radiationDamageModelBeta : XSDataDouble optional
# radiationDamageModelGamma : XSDataDouble optional
# rFriedel : XSDataDouble optional
# strategyOption : XSDataString optional
# transmission : XSDataDouble optional
# userDefinedRotationRange : XSDataAngle optional
# userDefinedRotationStart : XSDataAngle optional
# xdsBackgroundImage : XSDataFile optional
# xdsCorrectLp : XSDataFile optional
# xdsBkgpixCbf : XSDataFile optional
# xdsAsciiHkl : XSDataFile [] optional


class Best(AbstractTask):
    """
    This task runs the program BEST
    http://www.embl-hamburg.de/BEST
    """

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "diffractionPlan": {
                    "$ref": self.getSchemaUrl("ispybDiffractionPlan.json")
                },
                "subWedge": {
                    "type": "array",
                    "items": {"$ref": self.getSchemaUrl("subWedge.json")},
                },
                "xdsBackgroundImage": {"type": "string"},
                "correctLp": {"type": "string"},
                "xdsBkgpixCbf": {"type": "string"},
                "xdsAsciiHkl": {
                    "type": "array",
                    "items": {
                        "type": "string",
                    },
                },
            },
        }

    def run(self, inData):
        commandLine = self.createBestCommandLine(inData)
        self.runCommandLine(commandLine, listCommand=[])
        outData = {}
        return outData

    @staticmethod
    def createBestCommandLine(inData):
        diffractionPlan = inData.get("diffractionPlan", {})
        firstSubWedge = inData["subWedge"][0]
        experimentalCondition = firstSubWedge["experimentalCondition"]
        detector = experimentalCondition["detector"]
        beam = experimentalCondition["beam"]
        commandLine = "best"
        commandLine += Best.addOption(detector, "type", "-f")
        commandLine += Best.addOption(beam, "exposureTime", "-t")
        commandLine += Best.addOption(inData, "crystalAbsorbedDoseRate", "-GpS")
        commandLine += Best.addOption(diffractionPlan, "aimedCompleteness", "-C")
        commandLine += Best.addOption(diffractionPlan, "aimedIOverSigma", "-i2s")
        commandLine += Best.addOption(diffractionPlan, "aimedRedundancy", "-R")
        commandLine += Best.addOption(diffractionPlan, "aimedResolution", "-r")
        if diffractionPlan.get("anomalousData", False):
            if "numberOfCrystalPositions" in diffractionPlan:
                commandLine += " -a -p 0 360"
            elif "absorbedDoseRate" in inData:
                commandLine += " -asad"
            else:
                commandLine += " -a"
        commandLine += Best.addOption(diffractionPlan, "complexity", "-e")
        commandLine += Best.addOption(inData, "crystalShape", "-sh")
        commandLine += Best.addOption(diffractionPlan, "crystalSusceptibility", "-su")
        commandLine += Best.addOption(
            diffractionPlan, "detectorDistanceMax", "-DIS_MAX"
        )
        commandLine += Best.addOption(
            diffractionPlan, "detectorDistanceMin", "-DIS_MIN"
        )
        commandLine += Best.addOption(diffractionPlan, "doseLimit", "-DMAX")
        commandLine += Best.addOption(
            diffractionPlan, "goniostatMaxRotationSpeed", "-S"
        )
        commandLine += Best.addOption(
            diffractionPlan, "goniostatMinRotationWidth", "-w"
        )
        commandLine += Best.addOption(
            diffractionPlan, "maxExposureTimePerDataCollection", "-T"
        )
        commandLine += Best.addOption(diffractionPlan, "minExposureTimePerImage", "-M")
        commandLine += Best.addOption(diffractionPlan, "minTransmission", "-TRmin")
        commandLine += Best.addOption(
            diffractionPlan, "numberOfCrystalPositions", "-Npos"
        )
        commandLine += Best.addOption(
            diffractionPlan, "radiationDamageModelBeta", "-beta"
        )
        commandLine += Best.addOption(
            diffractionPlan, "radiationDamageModelGamma", "-gama"
        )
        commandLine += Best.addOption(diffractionPlan, "rFriedel", "-Rf")
        commandLine += Best.addOption(diffractionPlan, "strategyOption", "")
        commandLine += Best.addOption(diffractionPlan, "transmission", "-Trans")
        if (
            "userDefinedRotationRange" in diffractionPlan
            and "userDefinedRotationStart" in diffractionPlan
        ):
            commandLine += (
                " -phi {userDefinedRotationStart} {userDefinedRotationRange}".format(
                    **diffractionPlan
                )
            )
        # Output of GLE files and plotmtv files
        commandLine += " -g -o plot.mtv"
        # Integration data
        commandLine += " -xds " + inData["bkgpixCbf"]
        for xdsAsciiHklPath in inData["xdsAsciiHkl"]:
            commandLine += " " + xdsAsciiHklPath
        return commandLine

    @staticmethod
    def addOption(inData, name, option):
        returnValue = ""
        value = inData.get(name)
        if value is not None:
            returnValue = " {0} {1}".format(option, value)
        return returnValue



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "12/04/2021"

import math
import os
import fabio
import numpy
import time
import shutil
import scipy.ndimage

from PIL import Image
from PIL import ImageOps
from PIL import ImageFile
from PIL import ImageDraw
from PIL import ImageFont

from edna2.tasks.AbstractTask import AbstractTask
from edna2.tasks.ReadImageHeader import ReadImageHeader

from edna2.utils import UtilsPath
from edna2.utils import UtilsImage
from edna2.utils import UtilsConfig
from edna2.utils import UtilsLogging

logger = UtilsLogging.getLogger()


class DiffractionThumbnail(AbstractTask):
    """
    Generates diffraction thumbnail for PyArch
    """

    def getInDataSchema(self):
        return {
            "type": "object",
            "required": ["image"],
            "properties": {
                "image": {
                    "type": "array",
                    "items": {"type": "string"},
                },
                "forcedOutputDirectory": {"type": "string"},
                "waitForFileTimeOut": {"type": "number"},
                "format": {"type": "string"}
            }
        }

    def getOutDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "pathToJPEGImage": {
                    "type": "array",
                    "items": {"type": "string"}
                },
                "pathToThumbImage": {
                    "type": "array",
                    "items": {"type": "string"}
                },
            },
        }

    def run(self, inData):
        # Format suffix
        format = inData.get("format", "jpeg")
        if format == "jpeg":
            thumbSuffix = ".jpeg"
        elif format == "jpg":
            thumbSuffix = ".jpg"
        else:
            raise RuntimeError("Unsupported format: {0}".format(format))
        # Loop through all images
        listJPEGTask = []
        listThumbTask = []
        forcedOutputDirectory = inData.get("forcedOutputDirectory", None)
        for imagePath in inData["image"]:
            # Check image file extension
            imageFileName, suffix = os.path.splitext(os.path.basename(imagePath))
            if not suffix in [".img", ".marccd", ".mccd", ".cbf", ".h5"]:
                raise RuntimeError("Unknown image file name extension for pyarch thumbnail generator: %s" % imagePath)
            # Wait for image file
            if suffix == ".h5":
                h5MasterFilePath, h5DataFilePath, h5FileNumber = UtilsImage.getH5FilePath(imagePath, isFastMesh=True)
                waitFilePath = h5DataFilePath
            else:
                waitFilePath = imagePath
            expectedSize = self.getExpectedSize(imagePath)
            hasTimedOut, finalSize = UtilsPath.waitForFile(
                waitFilePath, expectedSize=expectedSize, timeOut=600)
            if hasTimedOut:
                raise RuntimeError("Waiting for file {0} timed out!".format(imagePath))
            outputFileName = imageFileName + thumbSuffix
            if forcedOutputDirectory is not None:
                outputPath = os.path.join(forcedOutputDirectory, outputFileName)
            else:
                outputPath = None
            # Create JPEG with resolution rings
            inDataReadHeader = {
                "imagePath": [imagePath],
                "skipNumberOfImages": True,
                "isFastMesh": True
            }
            readHeader = ReadImageHeader(
                inData=inDataReadHeader,
                workingDirectorySuffix=imageFileName,
            )
            readHeader.execute()
            experimentalCondition = readHeader.outData["subWedge"][0]["experimentalCondition"]
            detector = experimentalCondition["detector"]
            beam = experimentalCondition["beam"]
            inDataCreateJPEG = {
                "image": imagePath,
                "height": 1024,
                "width": 1024,
                "outputFileName": outputFileName,
                "outputPath": outputPath,
                "doResolutionRings": True,
                "pixelSizeX": detector["pixelSizeX"],
                "pixelSizeY": detector["pixelSizeY"],
                "beamPositionX": detector["beamPositionX"],
                "beamPositionY": detector["beamPositionY"],
                "distance": detector["distance"],
                "wavelength": beam["wavelength"],
            }
            createJPEG = CreateThumbnail(
                inData=inDataCreateJPEG,
                workingDirectorySuffix=imageFileName + "_JPEG"
            )
            createJPEG.start()
            listJPEGTask.append(createJPEG)
            # Create thumbnail
            outputFileName = imageFileName + ".thumb" + thumbSuffix
            if forcedOutputDirectory is not None:
                outputPath = os.path.join(forcedOutputDirectory, outputFileName)
            else:
                outputPath = None
            inDataCreateThumb = {
                "image": imagePath,
                "height": 256,
                "width": 256,
                "outputFileName": outputFileName,
                "outputPath": outputPath,
                "doResolutionRings": True,
                "pixelSizeX": detector["pixelSizeX"],
                "pixelSizeY": detector["pixelSizeY"],
                "beamPositionX": detector["beamPositionX"],
                "beamPositionY": detector["beamPositionY"],
                "distance": detector["distance"],
                "wavelength": beam["wavelength"],
            }
            createThumb = CreateThumbnail(
                inData=inDataCreateThumb,
                workingDirectorySuffix = imageFileName + "_thumbnail"
            )
            createThumb.start()
            listThumbTask.append(createThumb)
        outData = {
            "pathToJPEGImage": [],
            "pathToThumbImage": []
        }
        for task in listJPEGTask:
            task.join()
            if forcedOutputDirectory:
                outData["pathToJPEGImage"].append(task.outData["thumbNail"])
            else:
                pyarchPath = self.copyThumbnailToPyarch(task)
                outData["pathToJPEGImage"].append(pyarchPath)
        for task in listThumbTask:
            task.join()
            if forcedOutputDirectory:
                outData["pathToThumbImage"].append(task.outData["thumbNail"])
            else:
                pyarchPath = self.copyThumbnailToPyarch(task)
                outData["pathToThumbImage"].append(pyarchPath)
        return outData

    def getExpectedSize(self, imagePath):
        # Not great but works...
        expectedSize = 1000000
        for beamline in ["id23eh1", "id23eh2", "id30a1", "id30a2", "id30a3", "id30b"]:
            if beamline in imagePath:
                taskConfig = UtilsConfig.getTaskConfig("ExpectedFileSize", "esrf_" + beamline)
                expectedSize = int(taskConfig["image"])
                break
        return expectedSize

    def copyThumbnailToPyarch(self, task):
        imagePath = task.inData["image"]
        thumbNailPath = task.outData["thumbNail"]
        pyarchThumbnailDir = UtilsPath.createPyarchFilePath(os.path.dirname(imagePath))
        if pyarchThumbnailDir is None:
            pyarchThumbnailPath = thumbNailPath
        else:
            if not os.path.exists(pyarchThumbnailDir):
                os.makedirs(pyarchThumbnailDir, 0o755)
            pyarchThumbnailPath = os.path.join(
                pyarchThumbnailDir,
                os.path.basename(thumbNailPath)
            )
            shutil.copy2(thumbNailPath, pyarchThumbnailPath)
        return pyarchThumbnailPath

class CreateThumbnail(AbstractTask):

    def getInDataSchema(self):
        return {
            "type": "object",
            "required": ["image"],
            "properties": {
                "image": {"type": "string"},
                "height": {"type": "number"},
                "width": {"type": "number"},
                "outputFileName": {"type": "string"},
                "format": {"type": "string"},
                "doResolutionRings": {"type": "boolean"},
                "pixelSizeX": {"type": "number"},
                "pixelSizeY": {"type": "number"},
                "beamPositionX": {"type": "number"},
                "beamPositionY": {"type": "number"},
                "distance": {"type": "number"},
                "wavelength": {"type": "number"}
            }
        }

    def getOutDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "thumbNail": {"type": "string"}
            }
        }

    def run(self, inData):
        image = inData["image"]
        # Check if format is provided
        format = inData.get("format", None)
        height = inData.get("height", 512)
        width = inData.get("width", 512)
        outputPath = inData.get("outputPath", None)
        outputFileName = inData.get("outputFileName", None)
        doResolutionRings = inData.get("doResolutionRings", False)
        pixelSizeX = inData.get("pixelSizeX", False)
        pixelSizeY = inData.get("pixelSizeY", False)
        beamPositionX = inData.get("beamPositionX", False)
        beamPositionY = inData.get("beamPositionY", False)
        distance = inData.get("distance", False)
        wavelength = inData.get("wavelength", False)
        thumbNail = self.createThumbnail(
            image=image,
            format=format,
            height=height,
            width=width,
            outputPath=outputPath,
            workingDirectory=self.getWorkingDirectory(),
            outputFileName=outputFileName,
            doResolutionRings=doResolutionRings,
            pixelSizeX=pixelSizeX,
            pixelSizeY=pixelSizeY,
            beamPositionX=beamPositionX,
            beamPositionY=beamPositionY,
            distance=distance,
            wavelength=wavelength
        )
        outData = {
            "thumbNail": thumbNail
        }
        return outData


    @staticmethod
    def createThumbnail(image, format="jpg", height=512, width=512,
                        outputPath=None, minLevel=0, maxLevel=99.95,
                        dilatation=4, workingDirectory=None,
                        outputFileName=None, doResolutionRings=False,
                        pixelSizeX=None, pixelSizeY=None,
                        beamPositionX=None, beamPositionY=None,
                        distance=None, wavelength=None,
                        ):
        imageFileName = os.path.basename(image)
        imagePath = image
        imageNameWithoutSuffix, imageSuffix = os.path.splitext(imageFileName)
        if imageSuffix == ".h5":
            imageNumber = UtilsImage.getImageNumber(image)
            h5MasterFilePath, h5DataFilePath, h5FileNumber = UtilsImage.getH5FilePath(image, isFastMesh=True)
            noTrials = 5
            fabioImage = None
            while noTrials > 0:
                try:
                    fabioImage = fabio.openimage.openimage(h5MasterFilePath)
                    noTrials = 0
                except Exception as e:
                    logger.debug("Error when trying to open {0}: {1}".format(h5MasterFilePath, e))
                    logger.debug("Sleeping 5s and trying again, {0} trials left".format(noTrials))
                    noTrials -= 1
                    time.sleep(5)
            if fabioImage is None:
                raise RuntimeError("Cannot open file {0} with fabio".format(h5MasterFilePath))
            logger.debug(f"h5MasterFilePath: {h5MasterFilePath}")
            logger.debug(f"imageNumber: {imageNumber}, no frames: {fabioImage.nframes}")
            if imageNumber != 1 and imageNumber <= fabioImage.nframes:
                numpyImage = fabioImage.getframe(imageNumber-1).data
            else:
                numpyImage = fabioImage.data
            if numpyImage.dtype == numpy.dtype("uint32"):
                numpyImage = numpy.where(numpyImage > 65536*65536-2, 0, numpyImage)
            else:
                numpyImage = numpy.where(numpyImage > 256*256-2, 0, numpyImage)
        else:
            fabioImage = fabio.openimage.openimage(image)
            numpyImage = fabioImage.data
        # Default format
        suffix = "jpg"
        pilFormat = "JPEG"
        if format is not None:
            if format.lower() == "png":
                suffix = "png"
                pilFormat = "PNG"
        # The following code has been adapted from EDPluginExecThumbnail written by J.Kieffer
        dtype = numpyImage.dtype
        sortedArray = numpyImage.flatten()
        sortedArray.sort()
        numpyImage = numpy.maximum(numpyImage, int(minLevel) * numpy.ones_like(numpyImage))
        maxLevel = sortedArray[int(round(float(maxLevel) * sortedArray.size / 100.0))]
        if maxLevel < 25:
            maxLevel = 25
        numpyImage = numpy.minimum(numpyImage, maxLevel * numpy.ones_like(numpyImage))
        numpyImage = scipy.ndimage.morphology.grey_dilation(numpyImage, (dilatation, dilatation))
        mumpyImageFloat = (numpyImage.astype(numpy.float32)) / float(maxLevel)
        numpyImageInt = ( mumpyImageFloat * 255.0 ).astype(numpy.uint8)
        # Check if we should do resolution rings
        listResolution = []
        if doResolutionRings:
            delta = (height+width) / 2000
            if delta < 1.0:
                delta = 1.0
            centreX = beamPositionX / pixelSizeX
            centreY = beamPositionY / pixelSizeY
            sizeY, sizeX = numpyImageInt.shape
            averageSize = (sizeX + sizeY) / 2.0
            yy, xx = numpy.mgrid[:sizeY, :sizeX]
            circle = (xx - centreX) ** 2 + (yy - centreY) ** 2
            for resolution in [1.0, 1.1, 1.2, 1.5, 2.0, 3.0, 4.0]:
                if (wavelength/(2*resolution)) < 1.0:
                    theta = math.asin(wavelength/(2*resolution))
                    radius = math.tan(2*theta)* distance / pixelSizeX
                    listResolution.append((resolution, radius / averageSize ))
                    numpyImageInt = numpy.where(numpy.logical_and(circle < (radius+delta)**2, circle > (radius-delta)**2), 254, numpyImageInt)
        pilOutputImage = ImageOps.invert(Image.fromarray(numpyImageInt, 'L'))
        if height is not None and width is not None:
            pilOutputImage = pilOutputImage.resize((width, height), Image.ANTIALIAS)
        width, height = pilOutputImage.size
        for resolution, distance in listResolution:
            centreX = width / 2
            centreY = height / 2
            textfont = ImageFont.truetype("/usr/share/fonts/truetype/dejavu/DejaVuSans.ttf", int(height/30), encoding="unic")
            resolutionText = "{0} Å".format(resolution)
            imageEditable = ImageDraw.Draw(pilOutputImage)
            newDistance = distance  * (height + width) / 2.0 / math.sqrt(2)
            imageEditable.text((centreX + newDistance - width/20, centreY + newDistance -height/20), resolutionText, 0, font=textfont)
        if width * height > ImageFile.MAXBLOCK:
            ImageFile.MAXBLOCK = width * height
        if outputPath is None:
            if outputFileName is None:
                outputPath = os.path.join(workingDirectory, os.path.splitext(imageFileName)[0] + "." + suffix)
            else:
                outputPath = os.path.join(workingDirectory, outputFileName)
        pilOutputImage.save(outputPath, pilFormat, quality=85, optimize=True)
        logger.info("Output thumbnail path: %s" % outputPath)
        return outputPath



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "21/04/2019"

import os
import pathlib
import configparser


def getConfigDir():
    """
    Returns the directory used for configuration.
    If EDNA2_CONFIG is defined the value is returned,
    otherwise <project_base>/config.
    """
    if "EDNA2_CONFIG" in os.environ:
        configDir = pathlib.Path(os.environ["EDNA2_CONFIG"])
    else:
        pathFile = pathlib.Path(__file__)
        pathProjectBase = pathFile.parents[3]
        configDir = pathProjectBase / "config"
    return configDir


def getSite():
    """
    Returns the EDNA2_SITE variable from environment.
    If not defined returns 'Default'.
    """
    site = "Default"
    if "EDNA2_SITE" in os.environ:
        site = os.environ["EDNA2_SITE"]
    return site


def setSite(site):
    """
    Sets the EDNA2_SITE variable.
    """
    os.environ["EDNA2_SITE"] = site


def getConfig(site=None):
    config = configparser.ConfigParser()
    if site is None:
        site = getSite()
    configFile = site + ".ini"
    configDir = getConfigDir()
    configPath = configDir / configFile.lower()
    if configPath.exists():
        config.read(configPath.as_posix())
    return config


def getTaskConfig(taskName, site=None):
    dictConfig = {}
    config = getConfig(site)
    sections = config.sections()
    # First search in included configs
    if "Include" in sections:
        for site in config["Include"]:
            dictConfig.update(getTaskConfig(taskName, site))
    # Then update with the current config
    if taskName in sections:
        dictConfig.update(dict(config[taskName]))
    # Substitute ${} from os.environ
    for key in dictConfig:
        dictConfig[key] = os.path.expandvars(dictConfig[key])
    return dictConfig


def isEMBL():
    """
    Returns true if EMBL config
    """
    return getSite().lower().startswith("embl")


def isESRF():
    """
    Returns true if ESRF config
    """
    return getSite().lower().startswith("esrf")


def get(task, parameterName, defaultValue=None):
    if isinstance(task, str):
        taskConfig = getTaskConfig(task)
    else:
        taskConfig = getTaskConfig(task.__class__.__name__)
    value = taskConfig.get(parameterName.lower(), defaultValue)
    return value



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "10/05/2019"

# Corresponding EDNA code:
# https://github.com/olofsvensson/edna-mx
# mxPluginExec/plugins/EDPluginGroupMOSFLM-v1.0/plugins/EDPluginMOSFLMv10.py
# mxPluginExec/plugins/EDPluginGroupMOSFLM-v1.0/plugins/EDPluginMOSFLMIndexingv10.py

import os
import json

from edna2.tasks.AbstractTask import AbstractTask
from edna2.tasks.ReadImageHeader import ReadImageHeader

from edna2.utils import UtilsImage
from edna2.utils import UtilsConfig
from edna2.utils import UtilsDnaTables
from edna2.utils import UtilsLogging

logger = UtilsLogging.getLogger()


class AbstractMosflmTask(AbstractTask):
    """
    Common base class for all MOSLFM tasks
    """

    def __init__(self, inData):
        AbstractTask.__init__(self, inData)
        self.matrixFileName = None

    def run(self, inData):
        commandLine = 'ipmosflm DNA dnaTables.xml'
        # Translate from generic data model to specific MOSFLM data model
        mosflmInData = self.generateMOSFLMInData(inData)
        # Save MOSFLM inData
        with open(str(self.getWorkingDirectory() / "mosflmInData.json"), "w") as f:
            f.write(json.dumps(mosflmInData, indent=4))
        listCommand = self.generateMOSFLMCommands(mosflmInData, self.getWorkingDirectory())
        self.setLogFileName('mosflm.log')
        self.runCommandLine(commandLine, listCommand=listCommand)
        # Work in progress!
        outData = self.parseMosflmOutput(self.getWorkingDirectory())
        return outData

    @classmethod
    def generateMOSFLMInData(cls, inData):
        if "imagePath" in inData:
            inDataReadImageHeader = {
                "imagePath": inData["imagePath"]
            }
            readImageHeader = ReadImageHeader(inData=inDataReadImageHeader)
            readImageHeader.execute()
            listSubWedges = readImageHeader.outData["subWedge"]
        elif "subWedge" in inData:
            listSubWedges = inData["subWedge"]
        else:
            # Here we assume that we have inData in MOSFLM format...
            return inData
        firstSubWedge = listSubWedges[0]
        experimentalCondition = firstSubWedge["experimentalCondition"]
        detector = experimentalCondition["detector"]
        beam = experimentalCondition["beam"]
        goniostat = experimentalCondition["goniostat"]
        detectorType = detector["type"]
        if detectorType.startswith("pilatus"):
            mosflmDetector = "PILATUS"
        elif detectorType.startswith("eiger"):
            mosflmDetector = "EIGER"
        else:
            raise RuntimeError("Unknown detecor type for MOSFLM: " + detectorType)
        listImage = firstSubWedge["image"]
        firstImage = listImage[0]
        firstPath = firstImage["path"]
        mosflmInData = {
            # The MOSFLM beam position convention is opposite to the XDS / CBF/ Dozor convention
            "beam": {
                "x": detector["beamPositionY"],
                "y": detector["beamPositionX"]
            },
            "detector": {
                "numberPixelX": detector["numberPixelX"],
                "numberPixelY": detector["numberPixelY"],
                "pixelSizeX": detector["pixelSizeX"],
                "pixelSizeY": detector["pixelSizeY"],
                "type": mosflmDetector
            },
            "directory": os.path.dirname(firstPath),
            "distance": detector["distance"],
            "template": UtilsImage.getTemplate(firstPath),
            "wavelength": beam["wavelength"]
        }
        listMosflmImage = []
        for subWedge in listSubWedges:
            goniostat = subWedge["experimentalCondition"]["goniostat"]
            rotationAxisStart = goniostat['rotationAxisStart']
            oscillationWidth = goniostat['oscillationWidth']
            imageIndex = 0
            for image in subWedge["image"]:
                number = UtilsImage.getImageNumber(image["path"])
                rotationAxisStart = rotationAxisStart + oscillationWidth * imageIndex
                rotationAxisEnd = rotationAxisStart + oscillationWidth
                mosflmImage = {
                    "number":               number,
                    "rotationAxisStart":    rotationAxisStart,
                    "rotationAxisEnd":      rotationAxisEnd
                }
                listMosflmImage.append(mosflmImage)
        mosflmInData["image"] = listMosflmImage
        return mosflmInData
        #
        #                                                          "pixelSizeY"],
        #                                                      "image": [
        #     {
        #         "number": 1,
        #         "rotationAxisStart": 54.5,
        #         "rotationAxisEnd": 55.5
        #     },
        #     {
        #         "number": 2,
        #         "rotationAxisStart": 144.5,
        #         "rotationAxisEnd": 145.5
        #     }
        # ],
        # "matrix": {
        #     "matrixA": [
        #         [
        #             -0.00520639,
        #             0.00409075,
        #             0.00598777
        #         ],
        #         [
        #             -0.00922857,
        #             -0.00696139,
        #             0.00233786
        #         ],
        #         [
        #             0.00580022,
        #             -0.00738658,
        #             0.00518602
        #         ]
        #     ],
        #     "missettingsAngles": [
        #         0.0,
        #         0.0,
        #         0.0
        #     ],
        #     "matrixU": [
        #         [
        #             -0.4310086,
        #             0.3741439,
        #             0.8211259
        #         ],
        #         [
        #             -0.7639828,
        #             -0.6355417,
        #             -0.1114313
        #         ],
        #         [
        #             0.4801684,
        #             -0.6753541,
        #             0.5597635
        #         ]
        #     ],
        #     "cell": {
        #         "a": 74.4515,
        #         "b": 84.8724,
        #         "c": 115.476,
        #         "alpha": 70.0227,
        #         "beta": 76.0723,
        #         "gamma": 85.3234
        #     }
        # },
        # "mosaicityEstimation": 0.8500000000000001,
        # "deviationAngular": 1.906892,
        # "refinedDistance": 305.222015,
        # "symmetry": "P1"
        # }
        # outData = {
        #     "xds": xdsIndexingOutData
        # }

    def generateMOSFLMCommands(self, inData, workingDirectory):
        """
        This method creates a list of MOSFLM commands
        """
        detector = inData['detector']
        detectorType = detector['type']
        detectorCommand = detectorType
        # Check if reversephi and omega are configured
        omega = UtilsConfig.get('MOSFLM', 'omega')
        if omega is not None:
            detectorCommand += ' OMEGA {0}'.format(omega)
        reversePhi = UtilsConfig.get('MOSFLM', 'reversePhi')
        if reversePhi is not None:
            detectorCommand += ' REVERSEPHI'
        listCommand = [
            'WAVELENGTH {0}'.format(inData['wavelength']),
            'DISTANCE {0}'.format(inData['distance']),
            'BEAM {0} {1}'.format(inData['beam']['x'], inData['beam']['y']),
            'DIRECTORY {0}'.format(inData['directory']),
            'TEMPLATE {0}'.format(inData['template'])
        ]
        if 'symmetry' in inData:
            listCommand.append('SYMMETRY {0}'.format(inData['symmetry']))
        if 'mosaicity' in inData:
            listCommand.append('MOSAIC {0}'.format(inData['mosaicity']))

        newmatFileName = self.getNewmatFileName()
        newmatMatrix = inData.get('matrix', None)
        if newmatMatrix is not None:
            matrixFileName = self.getMatrixFileName()
            newmatPath = workingDirectory / matrixFileName
            self.writeNewmat(newmatMatrix, newmatPath)
            listCommand.append('MATRIX ' + matrixFileName)

        # Add exclude regions if Pilatus
        if detectorType == 'PILATUS':
            listCommand.append('DETECTOR {0}'.format(detectorCommand))
            if detector['numberPixelX'] == 1475 and \
               detector['numberPixelY'] == 1679:
                # Pilatus 2M
                listCommand.append('LIMITS EXCLUDE    0.0  83.76  288.96   85.14')
                listCommand.append('LIMITS EXCLUDE    0.0 168.73  288.96  170.10')
                listCommand.append('LIMITS EXCLUDE   33.54   0.0   36.63  253.87')
                listCommand.append('LIMITS EXCLUDE   70.00   0.0   73.1   253.87')
                listCommand.append('LIMITS EXCLUDE  106.46   0.0  109.56  253.87')
                listCommand.append('LIMITS EXCLUDE  142.93   0.0  146.02  253.87')
                listCommand.append('LIMITS EXCLUDE  179.39   0.0  182.49  253.87')
                listCommand.append('LIMITS EXCLUDE  215.86   0.0  218.95  253.87')
                listCommand.append('LIMITS EXCLUDE  252.32   0.0  255.42  253.87')
            elif detector['numberPixelX'] == 2463 and \
               detector['numberPixelY'] == 2527:
                # Pilatus 6M
                listCommand.append('LIMITS EXCLUDE  0.0 338.77 434.6 340.24')
                listCommand.append('LIMITS EXCLUDE  0.0 253.80 434.6 255.28')
                listCommand.append('LIMITS EXCLUDE  0.0 168.83 434.6 170.21')
                listCommand.append('LIMITS EXCLUDE  0.0  83.86 434.6  85.24')
                listCommand.append('LIMITS EXCLUDE 398.18  0.0 401.28 423.6')
                listCommand.append('LIMITS EXCLUDE 361.72  0.0 364.81 423.6')
                listCommand.append('LIMITS EXCLUDE 325.25  0.0 328.35 423.6')
                listCommand.append('LIMITS EXCLUDE 288.79  0.0 291.88 423.6')
                listCommand.append('LIMITS EXCLUDE 252.32  0.0 255.42 423.6')
                listCommand.append('LIMITS EXCLUDE 215.86  0.0 218.96 423.6')
                listCommand.append('LIMITS EXCLUDE 179.40  0.0 182.49 423.6')
                listCommand.append('LIMITS EXCLUDE 142.93  0.0 145.86 423.6')
                listCommand.append('LIMITS EXCLUDE 106.47  0.0 109.56 423.6')
                listCommand.append('LIMITS EXCLUDE  70.00  0.0 73.10  423.6')
                listCommand.append('LIMITS EXCLUDE  33.54  0.0 36.64  423.6')

        # Check if raster is configured
        raster = UtilsConfig.get('MOSFLM', 'raster')
        if raster is not None:
            listCommand.append('RASTER {0}'.format(raster))
        # Check if polarization is configured
        polarization = UtilsConfig.get('MOSFLM', 'polarization')
        if polarization is not None:
            listCommand.append('POLARIZATION {0}'.format(polarization))
        return listCommand

    def getNewmatFileName(self):
        newmatFileName = self.__class__.__name__ + '_newmat.mat'
        return newmatFileName

    def setNewmatFileName(self, newmatFileName):
        self.newmatFileName = newmatFileName

    def getMatrixFileName(self):
        if self.matrixFileName is None:
            self.matrixFileName = self.__class__.__name__ + '_matrix.mat'
        return self.matrixFileName

    def setMatrixFileName(self, matrixFileName):
        self.matrixFileName = matrixFileName

    @classmethod
    def splitStringIntoListOfFloats(cls, inputString):
        listFloats = []
        listString = inputString.split()
        for element in listString:
            if element != '':
                listFloats.append(float(element))
        return listFloats

    def getDataMOSFLMMatrix(self, matrixFileName):
        if matrixFileName is None:
            strMatrixFileName = self.getMatrixFileName()
        else:
            strMatrixFileName = matrixFileName
        xsDataMOSFLMNewmatMatrix = self.getNewmat(strMatrixFileName)
        return xsDataMOSFLMNewmatMatrix

    @classmethod
    def getNewmat(cls, newMatFilePath):
        listOfListOfFloat = []
        with open(str(newMatFilePath)) as f:
            listLine = f.readlines()
        # Convert into list of lists of float
        for line in listLine:
            if not line.startswith('SYMM'):
                listOfListOfFloat.append(
                    AbstractMosflmTask.splitStringIntoListOfFloats(line))
        # Fill in the data
        newmat = {}

        matrixA = [
            listOfListOfFloat[0],
            listOfListOfFloat[1],
            listOfListOfFloat[2],
        ]
        newmat['matrixA'] = matrixA
        newmat['missettingsAngles'] = listOfListOfFloat[3]
        newmat['matrixU'] = [
            listOfListOfFloat[4],
            listOfListOfFloat[5],
            listOfListOfFloat[6],
        ]
        cell = {
            'a': listOfListOfFloat[7][0],
            'b': listOfListOfFloat[7][1],
            'c': listOfListOfFloat[7][2],
            'alpha': listOfListOfFloat[7][3],
            'beta': listOfListOfFloat[7][4],
            'gamma': listOfListOfFloat[7][5]
        }
        newmat['cell'] = cell
        return newmat

    @classmethod
    def writeNewmat(cls, newmat, newmatPath):
        matrixA = newmat['matrixA']
        strNewmat = ''
        strNewmat += ' {0:11.8f} {1:11.8f} {2:11.8f}\n'.format(*matrixA[0])
        strNewmat += ' {0:11.8f} {1:11.8f} {2:11.8f}\n'.format(*matrixA[1])
        strNewmat += ' {0:11.8f} {1:11.8f} {2:11.8f}\n'.format(*matrixA[2])
        missettingsAngles = newmat['missettingsAngles']
        strNewmat += ' {0:11.3f} {0:11.3f} {0:11.3f}\n'.format(*missettingsAngles)
        matrixU = newmat['matrixU']
        strNewmat += ' {0:11.7f} {1:11.7f} {2:11.7f}\n'.format(*matrixU[0])
        strNewmat += ' {0:11.7f} {1:11.7f} {2:11.7f}\n'.format(*matrixU[1])
        strNewmat += ' {0:11.7f} {1:11.7f} {2:11.7f}\n'.format(*matrixU[2])
        cell = newmat['cell']
        strNewmat += ' {a:11.4f} {b:11.4f} {c:11.4f}'.format(**cell)
        strNewmat += ' {alpha:11.4f} {beta:11.4f} {gamma:11.4f}\n'.format(**cell)
        strNewmat += ' {0:11.3f} {0:11.3f} {0:11.3f}\n'.format(*missettingsAngles)
        with open(str(newmatPath), 'w') as f:
            f.write(strNewmat)

    # def generateExecutiveSummary(self, _edPlugin):
    #     """
    #     Generates a summary of the execution of the plugin.
    #     This method is common to all MOSFLM plugins.
    #     """
    #     self.DEBUG("EDPluginMOSFLMv10.generateExecutiveSummary")
    #     if (self.getStringVersion() is not None):
    #         self.addExecutiveSummaryLine(self.getStringVersion())


class MosflmIndexingTask(AbstractMosflmTask):

    def generateMOSFLMCommands(self, inData, workingDirectory):
        """
        This method creates a list of MOSFLM indexing commands given a valid
        inData
        """
        listCommand = AbstractMosflmTask.generateMOSFLMCommands(self,
                            inData, workingDirectory)
        listCommand.append('NEWMAT ' + self.getNewmatFileName())
        listImage = inData['image']
        for image in listImage:
            imageNumber = image['number']
            rotationAxisStart = image['rotationAxisStart']
            rotationAxisEnd = image['rotationAxisEnd']
            listCommand.append('AUTOINDEX DPS REFINE IMAGE {0} PHI {1} {2}'.format(
                imageNumber, rotationAxisStart, rotationAxisEnd))
        listCommand.append('GO')
        for image in listImage:
            listCommand.append('MOSAIC ESTIMATE {0}'.format(image['number']))
            listCommand.append('GO')
        return listCommand

    def parseMosflmOutput(self, workingDirectory):
        newMatFilePath = workingDirectory / self.getNewmatFileName()
        dnaTablesFilePath = workingDirectory / 'dnaTables.xml'
        return self.parseIndexingMosflmOutput(newMatFilePath, dnaTablesFilePath)

    def parseIndexingMosflmOutput(self, newMatFilePath, dnaTablesFilePath):
        outData = {}
        # Read Newmat file
        newmat = self.getNewmat(newMatFilePath)
        if newmat is None:
            errorMessage = "MOSFLM indexing error : No solution was obtained!"
            logger.error(errorMessage)
            self.setFailure()
        else:
            outData['newmat'] = self.getNewmat(newMatFilePath)
            # Then read the XML file
            dictDnaTables = UtilsDnaTables.getDict(dnaTablesFilePath)
            # Mosaicity estimation
            listTablesMosaicity = UtilsDnaTables.getTables(
                dictDnaTables, 'mosaicity_estimation')
            mosaicityValueSum = 0.0
            nValues = 0
            for table in listTablesMosaicity:
                listParam = UtilsDnaTables.getListParam(table)
                mosaicityValue = UtilsDnaTables.getListValue(
                    listParam, 'mosaicity', 'value')
                mosaicityValueSum += mosaicityValue
                nValues += 1
            outData['mosaicityEstimation'] = mosaicityValueSum / nValues
            # Refinement
            listTablesRefinement = UtilsDnaTables.getTables(
                dictDnaTables, 'refinement')
            tableRefinement = listTablesRefinement[0]
            listParamRefinement = UtilsDnaTables.getListParam(tableRefinement)
            outData['deviationAngular'] = UtilsDnaTables.getListValue(
                listParamRefinement, 'deviations', 'angular')
            outData['refinedDistance'] = UtilsDnaTables.getListValue(
                listParamRefinement, 'results', 'detector_distance')
            outData['spotsUsed'] = UtilsDnaTables.getListValue(
                listParamRefinement, 'parameters', 'used')
            outData['spotsTotal'] = UtilsDnaTables.getListValue(
                listParamRefinement, 'parameters', 'out_of')
            # Solution refinement
            listTablesSolutionRefinement = UtilsDnaTables.getTables(
                dictDnaTables, 'solution_refinement')
            tableSolutionRefinement = listTablesSolutionRefinement[0]
            listSolutionRefinement = UtilsDnaTables.getListParam(tableSolutionRefinement)
            outData['selectedSolutionNumber'] = UtilsDnaTables.getListValue(
                listSolutionRefinement, 'selection', 'number')
            outData['selectedSolutionSpaceGroup'] = UtilsDnaTables.getListValue(
                listSolutionRefinement, 'selection', 'spacegroup')
            outData['selectedSolutionSpaceGroupNumber'] = UtilsDnaTables.getListValue(
                listSolutionRefinement, 'selection', 'spacegroup_number')
            # Autoindex solutions
            listTablesAutoIndexSolution = UtilsDnaTables.getTables(
                dictDnaTables, 'autoindex_solutions')
            listIndexingSolution = []
            tablesAutoIndexSolution = listTablesAutoIndexSolution[0]
            listAutoindexSolutions = UtilsDnaTables.getListParam(tablesAutoIndexSolution)
            for autoIndexSolution in listAutoindexSolutions:
                indexingSolution = {
                    'index': UtilsDnaTables.getItemValue(
                        autoIndexSolution, 'index'),
                    'penalty': UtilsDnaTables.getItemValue(
                        autoIndexSolution, 'penalty'),
                    'lattice': UtilsDnaTables.getItemValue(
                        autoIndexSolution, 'lattice'),
                    'a': UtilsDnaTables.getItemValue(
                        autoIndexSolution, 'a'),
                    'b': UtilsDnaTables.getItemValue(
                        autoIndexSolution, 'b'),
                    'c': UtilsDnaTables.getItemValue(
                        autoIndexSolution, 'c'),
                    'alpha': UtilsDnaTables.getItemValue(
                        autoIndexSolution, 'alpha'),
                    'beta': UtilsDnaTables.getItemValue(
                        autoIndexSolution, 'beta'),
                    'gamma': UtilsDnaTables.getItemValue(
                        autoIndexSolution, 'gamma')
                }
                listIndexingSolution.append(indexingSolution)
            outData['indexingSolution'] = listIndexingSolution
            #  Beam refinement
            listTablesBeamRefinement = UtilsDnaTables.getTables(
                dictDnaTables, 'beam_refinement')
            tableBeamRefinement = listTablesBeamRefinement[0]
            listRefinedBeam = UtilsDnaTables.getListParam(tableBeamRefinement)
            initialBeamPositionX = UtilsDnaTables.getListValue(
                listRefinedBeam, 'initial_beam', 'x')
            initialBeamPositionY = UtilsDnaTables.getListValue(
                listRefinedBeam, 'initial_beam', 'y')
            refinedBeamPositionX = UtilsDnaTables.getListValue(
                listRefinedBeam, 'refined_beam', 'x')
            refinedBeamPositionY = UtilsDnaTables.getListValue(
                listRefinedBeam, 'refined_beam', 'y')
            beamPositionRefined = {
                'x': refinedBeamPositionX,
                'y': refinedBeamPositionY
            }
            outData['refinedBeam'] = beamPositionRefined
            beamPositionShift = {
                'x': initialBeamPositionX - refinedBeamPositionX,
                'y': initialBeamPositionY - refinedBeamPositionY,
            }
            outData['beamShift'] = beamPositionShift
        return outData
    #
    #
    # def generateExecutiveSummary(self, _edPlugin):
    #     """
    #     Generates a summary of the execution of the plugin.
    #     """
    #     EDPluginMOSFLMv10.generateExecutiveSummary(self, _edPlugin)
    #     self.DEBUG("EDPluginMOSFLMIndexingv10.createDataMOSFLMOutputIndexing")
    #     xsDataMOSFLMInputIndexing = self.getDataInput()
    #     xsDataMOSFLMOutputIndexing = self.getDataOutput()
    #     if not self.isFailure():
    #         self.addExecutiveSummaryLine("Execution of MOSFLM indexing successful.")
    #     self.addExecutiveSummaryLine("Image directory         : %s" % xsDataMOSFLMInputIndexing.getDirectory().getValue())
    #     self.addExecutiveSummaryLine("Image template          : %s" % xsDataMOSFLMInputIndexing.getTemplate().getValue())
    #     strImagesUsed = "Images used in indexing : "
    #     for xsDataMOSFLMImage in xsDataMOSFLMInputIndexing.getImage():
    #         strImagesUsed += "%5d" % xsDataMOSFLMImage.getNumber().getValue()
    #     self.addExecutiveSummaryLine(strImagesUsed)
    #     if (xsDataMOSFLMInputIndexing.getSymmetry() is not None):
    #         self.addExecutiveSummaryLine("Target symmetry     : %s" % xsDataMOSFLMInputIndexing.getSymmetry().getValue())
    #         self.addExecutiveSummaryLine("")
    #     if (xsDataMOSFLMOutputIndexing is not None):
    #         self.addExecutiveSummaryLine("")
    #         self.addExecutiveSummaryLine("List of possible solutions (index, penalty, lattice and cell):")
    #         iSelectedSolutionNumber = xsDataMOSFLMOutputIndexing.getSelectedSolutionNumber().getValue()
    #         # Add all solutions with penalty < 100 + 1 solution with penalty > 100
    #         bAddSolution = False
    #         listMOSFLMOutputIndexing = xsDataMOSFLMOutputIndexing.getPossibleSolutions()
    #         iNumberOfSolutions = len(listMOSFLMOutputIndexing)
    #         for i in range(iNumberOfSolutions):
    #
    #             iPenalty = listMOSFLMOutputIndexing[i].getPenalty().getValue()
    #             if (i < (iNumberOfSolutions - 1)):
    #                 iPenaltyNext = listMOSFLMOutputIndexing[i + 1].getPenalty().getValue()
    #                 if ((iPenalty >= 100) and (iPenaltyNext <= 100)):
    #                     bAddSolution = True
    #             if (bAddSolution):
    #                 iIndex = listMOSFLMOutputIndexing[i].getIndex().getValue()
    #                 xsDataCell = listMOSFLMOutputIndexing[i].getCell()
    #                 strLattice = listMOSFLMOutputIndexing[i].getLattice().getValue()
    #                 strPossibleSolution = "%3d %4d %2s %6.2f %6.2f %6.2f %6.2f %6.2f %6.2f" % \
    #                                             (iIndex, iPenalty, strLattice, \
    #                                               xsDataCell.getLength_a().getValue(),
    #                                               xsDataCell.getLength_b().getValue(),
    #                                               xsDataCell.getLength_c().getValue(),
    #                                               xsDataCell.getAngle_alpha().getValue(),
    #                                               xsDataCell.getAngle_beta().getValue(),
    #                                               xsDataCell.getAngle_gamma().getValue(),
    #                                              )
    #                 self.addExecutiveSummaryLine(strPossibleSolution)
    #             iPenaltyOld = iPenalty
    #         self.addExecutiveSummaryLine("")
    #         self.addExecutiveSummaryLine("Choosen solution number   : %14d" % iSelectedSolutionNumber)
    #         strSelectedSpaceGroup = xsDataMOSFLMOutputIndexing.getSelectedSolutionSpaceGroup().getValue()
    #         self.addExecutiveSummaryLine("Selected space group      : %14s" % (strSelectedSpaceGroup))
    #         xsDataCellRefined = xsDataMOSFLMOutputIndexing.getRefinedNewmat().getRefinedCell()
    #         self.addExecutiveSummaryLine("Refined cell              : %6.2f %7.2f %7.2f %5.1f %5.1f %5.1f" % (\
    #                                       xsDataCellRefined.getLength_a().getValue(),
    #                                       xsDataCellRefined.getLength_b().getValue(),
    #                                       xsDataCellRefined.getLength_c().getValue(),
    #                                       xsDataCellRefined.getAngle_alpha().getValue(),
    #                                       xsDataCellRefined.getAngle_beta().getValue(),
    #                                       xsDataCellRefined.getAngle_gamma().getValue()
    #                                       ))
    #         iSpotsTotal = xsDataMOSFLMOutputIndexing.getSpotsTotal().getValue()
    #         iSpotsUsed = xsDataMOSFLMOutputIndexing.getSpotsUsed().getValue()
    #         self.addExecutiveSummaryLine("Number of spots used      : %14d " % (iSpotsUsed))
    #         self.addExecutiveSummaryLine("Number of spots total     : %14d " % (iSpotsTotal))
    #         fDeviationPositional = xsDataMOSFLMOutputIndexing.getDeviationPositional().getValue()
    #         fDeviationAngular = xsDataMOSFLMOutputIndexing.getDeviationAngular().getValue()
    #         self.addExecutiveSummaryLine("Spot deviation positional : %14.2f [mm]" % (fDeviationPositional))
    #         self.addExecutiveSummaryLine("Spot deviation angular    : %14.2f [degrees]" % (fDeviationAngular))
    #         fBeamshiftX = xsDataMOSFLMOutputIndexing.getBeamShift().getX().getValue()
    #         fBeamshiftY = xsDataMOSFLMOutputIndexing.getBeamShift().getY().getValue()
    #         self.addExecutiveSummaryLine("Beam shift (X, Y)         : %6.3f, %6.3f [mm]" % \
    #                                       (fBeamshiftX, fBeamshiftY))
    #         fMosaicityEstimated = xsDataMOSFLMOutputIndexing.getMosaicityEstimation().getValue()
    #         self.addExecutiveSummaryLine("Estimated mosaicity       : %14.2f [degrees]" % fMosaicityEstimated)


class MosflmGeneratePredictionTask(AbstractMosflmTask):

    def __init__(self, inData):
        AbstractMosflmTask.__init__(self, inData)
        self.predictionFileName = None

    def generateMOSFLMCommands(self, inData, workingDirectory):
        """
        This method creates a list of MOSFLM generate prediction commands
        given a valid inData
        """
        listCommand = AbstractMosflmTask.generateMOSFLMCommands(self,
                                                                inData,
                                                                workingDirectory)
        template = inData['template']
        imageNumber = inData['image'][0]['number']
        self.predictionFileName = \
            MosflmGeneratePredictionTask.getImageFileNameFromTemplate(
                template, imageNumber)
        if self.predictionFileName is not None:
            self.predictionFileName += '_pred.jpg'
        rotationAxisStart = inData['image'][0]['rotationAxisStart']
        rotationAxisEnd = inData['image'][0]['rotationAxisEnd']
        listCommand.append('XGUI ON')
        listCommand.append('IMAGE %d PHI %f TO %f' % \
                           (imageNumber, rotationAxisStart, rotationAxisEnd))
        listCommand.append('GO')
        listCommand.append('PREDICT_SPOTS')
        listCommand.append(
            'CREATE_IMAGE PREDICTION ON BINARY TRUE FILENAME %s' % \
            self.predictionFileName)
        listCommand.append('RETURN')
        listCommand.append('EXIT')
        return listCommand

    @classmethod
    def getImageFileNameFromTemplate(cls, template, imageNumber):
        hashFound = False
        hasFinished = False
        firstHash = None
        noHashes = 0
        try:
            for index, character in enumerate(template):
                if (not hashFound) and (not hasFinished):
                    if character == '#':
                        firstHash = index
                        hashFound = True
                else:
                    if (character != '#') and (not hasFinished):
                        hasFinished = True
                if hashFound and (not hasFinished):
                    noHashes += 1
            imageFileName = template[0:firstHash] + \
                               str(imageNumber).rjust(noHashes, '0')
        except Exception:
            logger.warning("Couldn't transform template {}".format(template) +
                           " to file name")
            imageFileName = None
        return imageFileName

    def parseMosflmOutput(self, workingDirectory):
        predictionPath = None
        if self.predictionFileName is not None:
            predictionPath = workingDirectory / self.predictionFileName
        outData = {
            'predictionImage': predictionPath
        }
        return outData



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

__authors__ = "Olof Svensson, Igor Melnikov"
__contact__ = "svensson@esrf.fr"
__copyright__ = "ESRF"
__updated__ = "2021-07-20"

import os
import re

import numpy
import matplotlib
import matplotlib.cm
import matplotlib.pyplot as plt
import scipy

from edna2.tasks.AbstractTask import AbstractTask

from edna2.utils import UtilsImage
from edna2.utils import UtilsConfig
from edna2.utils import UtilsLogging
from edna2.utils import UtilsDetector


logger = UtilsLogging.getLogger()


class DozorM2(AbstractTask):  # pylint: disable=too-many-instance-attributes
    """
    The DozorM2 is responsible for executing the 'dozorm2' program.
    """

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "list_dozor_all": {
                    "type": "array",
                    "items": {"type": "string"},
                },
                "name_template_scan": {"type": "string"},
                "detectorType": {"type": "string"},
                "beamline": {"type": "string"},
                "detector_distance": {"type": "number"},
                "wavelength": {"type": "number"},
                "orgx": {"type": "number"},
                "orgy": {"type": "number"},
                "number_row": {"type": "number"},
                "number_images": {"type": "number"},
                "isZigZag": {"type": "boolean"},
                "step_h": {"type": "number"},
                "step_v": {"type": "number"},
                "beam_shape": {"type": "string"},
                "beam_h": {"type": "number"},
                "beam_v": {"type": "number"},
                "number_apertures": {"type": "integer"},
                "aperture_size": {"type": "string"},
                "reject_level": {"type": "integer"},
                "loop_thickness": {"type": "integer"},
                "isHorizontalScan": {"type": "boolean"},
                "number_scans": {"type": "integer"},
                "grid_x0": {"type": "number"},
                "grid_x1": {"type": "number"},
                "phi_values": {"type": "array", "items": {"type": "number"}},
                "sampx": {"type": "number"},
                "sampy": {"type": "number"},
                "phiy": {"type": "number"},
            },
        }

    def getOutDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "logPath": {"type": "string"},
                "workingDirectory": {"type": "string"},
            },
        }

    def run(self, inData):
        executable = UtilsConfig.get(self, "executable", "dozorm2")
        if len(inData["list_dozor_all"]) > 1:
            commandLine = f"{executable} -avs -cr dozorm2.dat"
        else:
            commandLine = f"{executable} dozorm2.dat"
        commands = self.generateCommands(inData, self.getWorkingDirectory())
        with open(str(self.getWorkingDirectory() / "dozorm2.dat"), "w") as f:
            f.write(commands)
        logPath = self.getWorkingDirectory() / "dozorm2.log"
        self.runCommandLine(commandLine, logPath=logPath)
        outData = self.parseOutput(self.getWorkingDirectory(), logPath)
        outData["logPath"] = str(logPath)
        outData["workingDirectory"] = str(self.getWorkingDirectory())
        return outData

    @staticmethod
    def generateCommands(inData, workingDirectory):
        """
        This method creates the input file for dozorm
        """
        detectorType = inData["detectorType"]
        nx = UtilsDetector.getNx(detectorType)
        ny = UtilsDetector.getNy(detectorType)
        pixelSize = UtilsDetector.getPixelsize(detectorType)
        for index, dozor_all_file in enumerate(inData["list_dozor_all"]):
            os.symlink(
                dozor_all_file, str(workingDirectory / "d_00{0}".format(index + 1))
            )
        nameTemplateScan = "d_00?"
        if inData.get("isHorizontalScan", True):
            meshDirect = "-h"
        else:
            meshDirect = "-v"
        command = "!\n"
        command += "detector {0}\n".format(detectorType)
        command += "nx %d\n" % nx
        command += "ny %d\n" % ny
        command += "pixel %f\n" % pixelSize
        command += "detector_distance {0}\n".format(inData["detector_distance"])
        command += "X-ray_wavelength {0}\n".format(inData["wavelength"])
        command += "orgx {0}\n".format(inData["orgx"])
        command += "orgy {0}\n".format(inData["orgy"])
        command += "number_row {0}\n".format(inData["number_row"])
        command += "number_images {0}\n".format(inData["number_images"])
        command += "mesh_direct {0}\n".format(meshDirect)
        command += "step_h {0}\n".format(inData["step_h"])
        command += "step_v {0}\n".format(inData["step_v"])
        command += "beam_shape {0}\n".format(inData["beam_shape"])
        command += "beam_h {0}\n".format(inData["beam_h"])
        command += "beam_v {0}\n".format(inData["beam_v"])
        command += "number_apertures {0}\n".format(inData["number_apertures"])
        command += "aperture_size {0}\n".format(inData["aperture_size"])
        command += "reject_level {0}\n".format(inData["reject_level"])
        if "loop_thickness" in inData and inData["loop_thickness"] is not None:
            command += "loop_thickness {0}\n".format(inData["loop_thickness"])
        command += "name_template_scan {0}\n".format(nameTemplateScan)
        command += "number_scans {0}\n".format(inData["number_scans"])
        command += "first_scan_number 1\n"
        if "phi_values" in inData:
            for index, phi_value in enumerate(inData["phi_values"]):
                command += "phi{0} {1}\n".format(index + 1, phi_value)
            command += "axis_zero {0} {1}\n".format(
                inData["grid_x0"], inData["grid_y0"]
            )
        if "sampx" in inData:
            command += "sampx {0}\n".format(inData["sampx"])
        if "sampy" in inData:
            command += "sampy {0}\n".format(inData["sampy"])
        if "phiy" in inData:
            command += "phiy {0}\n".format(inData["phiy"])
        command += "end\n"
        # logger.debug('command: {0}'.format(command))
        return command

    @staticmethod
    def parseOutput(working_dir, log_path):
        outData = {}
        path_dozorm2_map = working_dir / "dozorm_001.map"
        if path_dozorm2_map.exists():
            dict_map = DozorM2.parseMap(path_dozorm2_map)
            nx = dict_map["nx"]
            ny = dict_map["ny"]
            dict_coord = DozorM2.parseDozorm2LogFile(log_path)
            crystal_map_path = DozorM2.makeCrystalPlot(dict_map["crystal"], working_dir)
            if nx != 1 and ny != 1:
                image_number_map_path = DozorM2.makeImageNumberMap(
                    dict_map["imageNumber"], working_dir
                )
            else:
                image_number_map_path = None
            # Colour map
            D, Z = DozorM2.parser(path_dozorm2_map)
            colour_map_path = DozorM2.colourMapPlot(Z, D, working_dir=working_dir)

            outData = {
                "dozorMap": str(path_dozorm2_map),
                "dictCoord": dict_coord,
                "nx": nx,
                "ny": ny,
                "score": dict_map["score"],
                "crystal": dict_map["crystal"],
                "imageNumber": dict_map["imageNumber"],
                "crystalMapPath": crystal_map_path,
                "colourMapPath": colour_map_path,
                "imageNumberMapPath": image_number_map_path,
            }
        return outData

    @staticmethod
    def parseDozorm2LogFile(logPath):
        #                    SCAN 1
        #                    ======
        #
        #    Total N.of crystals in Loop =  3
        # Cryst Aperture Central  Coordinate  Int/Sig  N.of Images CRsize Score   Dmin Helic   Start     Finish     Int/Sig
        # number size     image      X    Y          All  dX  dY   X   Y   sum                 x     y     x     y   helical
        # ------------------------------------------------------------------------------------------------------------------> X
        #     1   100.0     125   28.0    4.2  172.1  47  13   5  12   4  3846.2  3.06   NO
        #     2    20.0     133   20.0    4.0   47.5   5   3   2   3   2   147.8  3.46  YES    18     4    20     4   198.0
        #     3    20.0     198   31.0    6.0   37.5   2   2   1   2   1   112.2  3.68   NO
        with open(str(logPath)) as fd:
            listLogLines = fd.readlines()
        doParseLine = False
        do3dCoordinates = False
        scan1 = None
        scan2 = None
        listCoord = None
        for line in listLogLines:
            # print([line])
            if "SCAN 1" in line:
                listPositions = []
            elif "SCAN 2" in line:
                scan1 = listPositions
                listPositions = []
            elif "3D COORDINATES" in line:
                scan2 = listPositions
                do3dCoordinates = True
                listCoord = []
            if line.startswith("------"):
                doParseLine = True
            elif len(line) == 1:
                doParseLine = False
            elif doParseLine:
                listValues = line.split()
                if not do3dCoordinates:
                    try:
                        iOverSigma = float(listValues[5])
                    except ValueError:
                        iOverSigma = listValues[5]
                    position = {
                        "number": int(listValues[0]),
                        "apertureSize": str(int(float(listValues[1]))),
                        "imageNumber": int(listValues[2]),
                        "xPosition": float(listValues[3]),
                        "yPosition": float(listValues[4]),
                        "iOverSigma": iOverSigma,
                        "numberOfImagesTotal": int(listValues[6]),
                        "numberOfImagesTotalX": int(listValues[7]),
                        "numberOfImagesTotalY": int(listValues[8]),
                        "crSizeX": int(listValues[9]),
                        "crSizeY": int(listValues[10]),
                        "score": float(listValues[11]),
                        "dmin": float(listValues[12]),
                        "helical": listValues[13] == "YES",
                    }
                    if position["helical"]:
                        position["helicalStartX"] = listValues[14]
                        position["helicalStartY"] = listValues[15]
                        position["helicalStopX"] = listValues[16]
                        position["helicalStopY"] = listValues[17]
                        position["helicalIoverSigma"] = listValues[18]
                    listPositions.append(position)
                else:
                    coord = {
                        "number": int(listValues[0]),
                        "averageScore": float(listValues[1]),
                        "dmin": float(listValues[2]),
                        "sc1": int(listValues[3]),
                        "sc2": int(listValues[4]),
                        "size": float(listValues[5]),
                        "scanX": float(listValues[6]),
                        "scanY1": float(listValues[7]),
                        "scanY2": float(listValues[8]),
                        "dx": float(listValues[9]),
                        "dy1": float(listValues[10]),
                        "dy2": float(listValues[11]),
                        "sampx": float(listValues[12]),
                        "sampy": float(listValues[13]),
                        "phiy": float(listValues[14]),
                        # "alfa": float(listValues[15]),
                        # "sampx": float(listValues[16]),
                        # "sampy": float(listValues[17]),
                        # "phiy": float(listValues[18])
                    }
                    listCoord.append(coord)
        if scan1 is None:
            scan1 = listPositions
        dictCoord = {"scan1": scan1, "scan2": scan2, "coord": listCoord}
        return dictCoord

    @staticmethod
    def makeCrystalPlot(arrayCrystal, workingDirectory, debug=False):
        npArrayCrystal = numpy.array(arrayCrystal)
        ySize, xSize = npArrayCrystal.shape
        if xSize == 1:
            # Vertical line scan - transpose the matrix
            npArrayCrystal = numpy.transpose(npArrayCrystal)
            ySize, xSize = npArrayCrystal.shape
        npArrayCrystalAbs = numpy.abs(npArrayCrystal)
        # Make '999' be the max crystal number + 1
        maxNumber = numpy.amax(
            numpy.where(npArrayCrystalAbs < 999, npArrayCrystalAbs, 0)
        )
        npArrayCrystalAbs = numpy.where(
            npArrayCrystalAbs == 999, maxNumber + 1, npArrayCrystalAbs
        )
        # minValue = numpy.amin(npArrayCrystal)
        # newZeroValue = minValue - 1
        # npArrayCrystal = numpy.where(npArrayCrystal == 0.0, newZeroValue, npArrayCrystal)

        maxSize = max(xSize, ySize)
        if maxSize < 10:
            fontSize = 12
            dpi = 75
        elif maxSize < 50:
            fontSize = 8
            dpi = 100
        else:
            fontSize = 5
            dpi = 150

        font = {"family": "normal", "weight": "normal", "size": fontSize}

        matplotlib.rc("font", **font)

        fig, ax = plt.subplots()

        _ = ax.imshow(npArrayCrystalAbs, cmap=matplotlib.cm.Spectral)

        ax.set_xticks(numpy.arange(len(range(xSize))))
        ax.set_yticks(numpy.arange(len(range(ySize))))

        ax.set_xticklabels(list(range(1, xSize + 1)))
        ax.set_yticklabels(list(range(1, ySize + 1)))

        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

        # Loop over data dimensions and create text annotations.
        for i in range(ySize):
            for j in range(xSize):
                if abs(npArrayCrystal[i, j]) > 0.001:
                    _ = ax.text(
                        j, i, npArrayCrystal[i, j], ha="center", va="center", color="b"
                    )

        ax.set_title("Crystal map")
        fig.tight_layout(pad=2)
        w, h = fig.get_size_inches()
        x1, x2 = ax.get_xlim()
        y1, y2 = ax.get_ylim()
        x1 = float(x1)
        x2 = float(x2)
        y1 = float(y1)
        y2 = float(y2)
        fig.set_size_inches(w + 2, abs(y2 - y1) / (x2 - x1) * w + 2)
        if debug:
            plt.show()
        crystalMapPath = os.path.join(workingDirectory, "crystalMap.png")
        plt.savefig(crystalMapPath, dpi=dpi)
        UtilsImage.crop_image(crystalMapPath)

        return crystalMapPath

    @staticmethod
    def makeImageNumberMap(arrayImageNumber, workingDirectory, debug=False):
        npImageNumber = numpy.array(arrayImageNumber)
        npArrayImageNumber = numpy.zeros(npImageNumber.shape)
        ySize, xSize = npImageNumber.shape

        maxSize = max(xSize, ySize)
        if maxSize < 10:
            fontSize = 12
            dpi = 75
        elif maxSize < 50:
            fontSize = 8
            dpi = 100
        else:
            fontSize = 5
            dpi = 150

        font = {"family": "normal", "weight": "normal", "size": fontSize}

        matplotlib.rc("font", **font)

        fig, ax = plt.subplots()
        _ = ax.imshow(npArrayImageNumber, cmap=matplotlib.cm.Greys)

        ax.set_xticks(numpy.arange(len(range(xSize))))
        ax.set_yticks(numpy.arange(len(range(ySize))))

        ax.set_xticklabels(list(range(1, xSize + 1)))
        ax.set_yticklabels(list(range(1, ySize + 1)))

        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

        # Loop over data dimensions and create text annotations.
        for i in range(ySize):
            for j in range(xSize):
                _ = ax.text(
                    j, i, arrayImageNumber[i][j], ha="center", va="center", color="b"
                )

        ax.set_title("Image numbers")
        fig.tight_layout(pad=2)
        w, h = fig.get_size_inches()
        x1, x2 = ax.get_xlim()
        y1, y2 = ax.get_ylim()
        x1 = float(x1)
        x2 = float(x2)
        y1 = float(y1)
        y2 = float(y2)
        fig.set_size_inches(w + 2, abs(y2 - y1) / (x2 - x1) * w + 2)
        if debug:
            plt.show()
        imageNumberPath = os.path.join(workingDirectory, "imageNumber.png")
        plt.savefig(imageNumberPath, dpi=dpi)
        UtilsImage.crop_image(imageNumberPath)

        return imageNumberPath

    @staticmethod
    def parseMatrix(index, listLines, spacing, isFloat=True):
        arrayValues = []
        # Parse matrix - starts and ends with "---" line
        while not listLines[index].startswith("-------"):
            index += 1
        index += 1
        while not listLines[index].startswith("-------"):
            # listScores = textwrap.wrap(listLines[index][5:], spacing)
            listScores = []
            for line_pos in range(0, len(listLines[index]), spacing):
                sub_string = listLines[index][
                    line_pos + 5 : line_pos + spacing + 5
                ].strip()
                if sub_string != "":
                    if isFloat:
                        listScores.append(float(sub_string))
                    else:
                        listScores.append(int(sub_string))
            arrayValues.append(listScores)
            index += 1
        index += 1
        return index, arrayValues

    @staticmethod
    def parseMap(mapPath):
        with open(str(mapPath)) as fd:
            listLines = fd.readlines()
        # Parse map dimensions
        index = 1
        nx, ny = map(int, listLines[index].split())
        # Parse scores
        index, arrayScore = DozorM2.parseMatrix(
            index, listLines, spacing=6, isFloat=True
        )
        # Parse rel. contamination
        index, relContamination = DozorM2.parseMatrix(
            index, listLines, spacing=6, isFloat=True
        )
        # Parse crystals
        index, arrayCrystal = DozorM2.parseMatrix(
            index, listLines, spacing=4, isFloat=False
        )
        # Parse image number
        index, arrayImageNumber = DozorM2.parseMatrix(
            index, listLines, spacing=5, isFloat=False
        )
        dictMap = {
            "nx": nx,
            "ny": ny,
            "score": arrayScore,
            "relContamination": relContamination,
            "crystal": arrayCrystal,
            "imageNumber": arrayImageNumber,
        }
        return dictMap

    @staticmethod
    def updateMeshPositions(meshPositions, arrayScore):
        newMeshPositions = []
        for position in meshPositions:
            # pprint.pprint(position)
            indexY = position["indexY"]
            indexZ = position["indexZ"]
            # print(indexY, indexZ)
            dozormScore = arrayScore[indexZ][indexY]
            dozorScore = position["dozor_score"]
            # print(dozorScore, dozormScore)
            newPosition = dict(position)
            newPosition["dozor_score_orig"] = dozorScore
            newPosition["dozor_score"] = dozormScore
            newMeshPositions.append(newPosition)
        return newMeshPositions

    @staticmethod
    def check1Dpositions(listPositions, nx, ny):
        newListPositions = []
        for position in listPositions:
            newPosition = dict(position)
            if nx == 1:
                newPosition["xPosition"] = 1.0
            if ny == 1:
                newPosition["yPosition"] = 1.0
            newListPositions.append(newPosition)
        return newListPositions

    @staticmethod
    def constructColorlist(array):
        basecolors = [
            "#00CA02",
            "#FF0101",
            "#F5A26F",
            "#668DE5",
            "#E224DE",
            "#04FEFD",
            "#FEFE00",
            "#0004AF",
            "#B5FF06",
        ]

        N = int(numpy.max(array))
        AdjacentArray = numpy.identity(N)

        t = numpy.ones((3, 3), dtype="int32")
        for j in range(1, N + 1):
            cut = scipy.ndimage.measurements.label(array == j)
            c = scipy.signal.convolve2d(cut[0], t, mode="same")
            adjacentvalues = numpy.unique(array[numpy.where(c != 0)]).astype("int")

            for i in adjacentvalues:
                if i == -1 or i == -2:
                    pass
                else:
                    AdjacentArray[j - 1, i - 1] = 1

        #    print(AdjacentArray)

        t = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        Adjacent_NULL = numpy.tril(AdjacentArray, k=-1)
        AdjacentArray = Adjacent_NULL

        ColorVector = numpy.ones(N)
        for i in range(N):
            BannedColors = numpy.unique(AdjacentArray[i, :])[1:]
            for item in BannedColors:
                t.remove(item)
            ColorVector[i] = t[0]
            AdjacentArray = ColorVector * Adjacent_NULL
            t = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        ColorVector = ColorVector.astype(int)

        #    random.shuffle(basecolors)
        colors = ColorVector.astype(str)
        for i in range(N):
            colors[i] = basecolors[ColorVector[i] - 1]

        clrs = ["grey", "black", "black"]
        clrs.extend(colors.tolist())

        #    print(clrs)

        return clrs

    @staticmethod
    def colourMapPlot(crystalN_array, dozorscore_array, working_dir=None):
        fig, ax = plt.subplots()

        Ztable, Dtable = crystalN_array, dozorscore_array
        row, col = Dtable.shape

        Zcopy = numpy.copy(Ztable)
        Ztable = numpy.abs(Ztable)
        Ztable[Ztable == 999] = -2

        #    print(numpy.sort(numpy.unique(Ztable[Ztable>0])))

        clrs = DozorM2.constructColorlist(Ztable)
        cmap = matplotlib.colors.ListedColormap(
            [matplotlib.colors.to_rgba(str(i)) for i in clrs]
        )
        bounds = numpy.arange(-2.5, int(numpy.max(Ztable)) + 1)
        norm = matplotlib.colors.BoundaryNorm(bounds, len(clrs))
        if numpy.max(Ztable > 0):
            plt.imshow(
                Ztable,
                cmap=cmap,
                norm=norm,
                interpolation="nearest",
                origin="upper",
                extent=[0.5, (col + 0.5), (row + 0.5), 0.5],
            )

            m = int(numpy.log10(numpy.max(Dtable)))
            if m < 1:
                m = 1
            M = numpy.max(Dtable)
            for (j, i) in numpy.ndindex((row, col)):
                if Ztable[j, i] > 0:
                    if (j, i + 1) in numpy.ndindex((row, col)):
                        if Ztable[j, i + 1] != Ztable[j, i]:
                            line = plt.Line2D(
                                (i + 1.5, i + 1.5),
                                (j + 0.5, j + 1.5),
                                lw=2,
                                color="white",
                            )
                            plt.gca().add_line(line)
                    if (j, i - 1) in numpy.ndindex((row, col)):
                        if Ztable[j, i - 1] != Ztable[j, i]:
                            line = plt.Line2D(
                                (i + 0.5, i + 0.5),
                                (j + 0.5, j + 1.5),
                                lw=2,
                                color="white",
                            )
                            plt.gca().add_line(line)
                    if (j + 1, i) in numpy.ndindex((row, col)):
                        if Ztable[j + 1, i] != Ztable[j, i]:
                            line = plt.Line2D(
                                (i + 0.5, i + 1.5),
                                (j + 1.5, j + 1.5),
                                lw=2,
                                color="white",
                            )
                            plt.gca().add_line(line)
                    if (j - 1, i) in numpy.ndindex((row, col)):
                        if Ztable[j - 1, i] != Ztable[j, i]:
                            line = plt.Line2D(
                                (i + 0.5, i + 1.5),
                                (j + 0.5, j + 0.5),
                                lw=2,
                                color="white",
                            )
                            plt.gca().add_line(line)

                    plt.text(
                        i + 1,
                        j + 1,
                        Zcopy[j, i].astype(int),
                        c="black",
                        ha="center",
                        va="center",
                        size=3,
                    )
                    ax.add_patch(
                        matplotlib.patches.Circle(
                            (i + 1, j + 1), Dtable[j, i] / (3 * M), color="white"
                        )
                    )
                elif Ztable[j, i] == -2:
                    ax.add_patch(
                        matplotlib.patches.Circle(
                            (i + 1, j + 1), Dtable[j, i] / (3 * M), color="white"
                        )
                    )

            plt.text(
                col + 1, 2, "Dozor\nscore", c="black", ha="left", va="center", size=10
            )
            ax.add_patch(
                matplotlib.patches.Circle(
                    (col + 1.5, 4),
                    0.333 * round(M, -m + 1) / M,
                    color="black",
                    clip_on=False,
                )
            )
            plt.text(
                col + 2,
                4,
                ("{:." + ("0" if m > 0 else "1") + "f}").format(round(M, -m + 1)),
                c="black",
                ha="left",
                va="center",
                size=10,
            )
        else:
            plt.imshow(
                Dtable,
                cmap="hot",
                interpolation="nearest",
                origin="upper",
                extent=[0.5, (col + 0.5), (row + 0.5), 0.5],
            )

        plt.xticks(numpy.arange(1, Ztable.shape[1] + 1, 2), rotation=45)
        plt.yticks(numpy.arange(1, Ztable.shape[0] + 1, 2))
        plot_name = "colourMap.png"
        if working_dir is None:
            crystal_map_path = os.path.join(os.getcwd(), plot_name)
        else:
            crystal_map_path = os.path.join(working_dir, plot_name)
        plt.savefig(crystal_map_path, dpi=200)
        UtilsImage.crop_image(crystal_map_path)
        # plt.show()
        plt.close()
        return crystal_map_path

    @staticmethod
    def parser(filename):
        l = []
        with open(filename, "r") as f:
            l = f.readlines()
            f.close()

        #    print(re.split(r'\s+', l[1])[1:3])
        col, row = [int(x) for x in re.split(r"\s+", l[1])[1:3]]

        #    Dtable
        Dtable = numpy.zeros((row, col))
        i = -4
        for line in l:
            if re.search("Map of Scores", line):
                i = -3

            if i >= 0:
                #            print(numpy.array(re.findall('.{6}', line[5:])).astype(float).size)
                Dtable[i, :] = numpy.array(re.findall(".{6}", line[5:])).astype(float)

            if i >= row - 1:
                break

            if i >= -3:
                i += 1

        #    Ztable
        Ztable = numpy.zeros((row, col))
        i = -4
        for line in l:
            if re.search("Map of Crystals", line):
                i = -3

            if i >= 0:
                #            print(numpy.array(re.findall('.{4}', line[5:])).astype(int).size)
                Ztable[i, :] = numpy.array(re.findall(".{4}", line[5:])).astype(int)

            if i >= row - 1:
                break

            if i >= -3:
                i += 1

        # plt.imshow(Dtable)
        # plt.show()
        #
        return Dtable, Ztable



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "21/04/2019"



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the 'Software'), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ['O. Svensson']
__license__ = 'MIT'
__date__ = '25/03/2021'

import os
import fabio
import pathlib

from edna2.tasks.AbstractTask import AbstractTask
from edna2.tasks.ReadImageHeader import ReadImageHeader

from edna2.utils import UtilsImage
from edna2.utils import UtilsConfig
from edna2.utils import UtilsLogging

logger = UtilsLogging.getLogger()


class H5ToBinnedCBFTask(AbstractTask):

    def getInDataSchema(self):
        return {
            "type": "object",
            "required": ["hdf5File"],
            "properties": {
                "imageNumber": {"type": "integer"},
                "startImageNumber": {"type": "integer"},
                "hdf5ImageNumber": {"type": "integer"},
                "hdf5File": {"type": "string"},
                "forcedOutputDirectory": {"type": "string"}
            }
        }

    def getOutDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "outputCBFFile": {"type": "string"}
            }
        }

    def run(self, inData):
        outData = {}
        hdf5File = pathlib.Path(inData['hdf5File'])
        # Read the header
        inDataReadImageHeader = {
            "imagePath": [inData['hdf5File']],
            "isFastMesh": False
        }
        readImageHeader = ReadImageHeader(
            inData = inDataReadImageHeader
        )
        readImageHeader.execute()
        if readImageHeader.isSuccess():
            firstSubWedge = readImageHeader.outData["subWedge"][0]
            experimentalCondition = firstSubWedge["experimentalCondition"]
            detector = experimentalCondition["detector"]
            beam = experimentalCondition["beam"]
            goniostat = experimentalCondition["goniostat"]
            beamX = detector["beamPositionX"] / 2.0 / detector["pixelSizeX"]
            beamY = detector["beamPositionY"] / 2.0 / detector["pixelSizeY"]
            directory = hdf5File.parent
            prefix = UtilsImage.getPrefix(hdf5File)
            hdf5ImageNumber = 1
            if 'master.h5' in str(hdf5File):
                masterFile = hdf5File
            else:
                if UtilsConfig.isEMBL():
                    fileName = '{0}_master.h5'.format(prefix)
                else:
                    fileName = '{0}_{1}_master.h5'.format(prefix, hdf5ImageNumber)
                masterFile = directory / fileName
            image = fabio.open(str(masterFile))
            image.data = image.data.reshape(2181, 2, 2074, 2).sum(3).sum(1)
            cbfImage = image.convert("cbf")
            pilatus_headers = fabio.cbfimage.PilatusHeader("Silicon sensor, thickness 0.000750 m")
            pilatus_headers["Start_angle"] = goniostat["rotationAxisStart"]
            pilatus_headers["Angle_increment"] = goniostat["oscillationWidth"]
            pilatus_headers["Detector"] = "Eiger2 16M binned to 4M"
            pilatus_headers["Pixel_size"] = (detector["pixelSizeY"]*2, detector["pixelSizeX"]*2)
            pilatus_headers["Exposure_time"] = beam["exposureTime"]
            pilatus_headers["Wavelength"] = beam["wavelength"]
            pilatus_headers["Detector_distance"] = detector["distance"] / 1000.0
            pilatus_headers["Beam_xy"] = (beamY, beamX)
            pilatus_headers["Count_cutoff"] = 1009869
            cbfImage.pilatus_headers = pilatus_headers
            directory = inData.get("forcedOutputDirectory", str(self.getWorkingDirectory()))
            cbfImagePath = os.path.join(directory, os.path.basename(inData['hdf5File']).replace(".h5", ".cbf"))
            cbfImage.save(cbfImagePath)
            outData["outputCBFFile"] = cbfImagePath
        return outData



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "14/04/2020"

import os
import numpy as np

from edna2.tasks.AbstractTask import AbstractTask
from edna2.tasks.SubWedgeAssembly import SubWedgeAssembly
from edna2.tasks.ControlDozor import ControlDozor
from edna2.tasks.XDSTasks import XDSIndexing
from edna2.utils import UtilsImage


class ControlIndexing(AbstractTask):
    """
    This task receives a list of images or data collection ids and
    returns result of indexing
    """

    def getInDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "dataCollectionId": {"type": "integer"},
                "imagePath": {
                    "type": "array",
                    "items": {
                        "type": "string",
                    },
                },
            },
        }

    def run(self, in_data):
        outData = {}
        # First get the list of subWedges
        if "subWedge" in in_data:
            list_sub_wedge = in_data["subWedge"]
        else:
            list_sub_wedge = self.getListSubWedge(in_data)
        # # Get list of spots from Dozor
        # listOutDataControlDozor = self.runControlDozor(listSubWedge)
        # listDozorSpotFile = []
        # for outDataControlDozor in listOutDataControlDozor:
        #     if "dozorSpotFile" in outDataControlDozor["imageQualityIndicators"][0]:
        #         dozorSpotFile = outDataControlDozor["imageQualityIndicators"][0][
        #             "dozorSpotFile"
        #         ]
        #         listDozorSpotFile.append(dozorSpotFile)
        # imageDict = listSubWedge[0]
        # Run XDS indexing
        xds_indexin_in_data = {
            "subWedge": list_sub_wedge
            # "dozorSpotFile": listDozorSpotFile,
        }
        xds_indexing_task = XDSIndexing(
            inData=xds_indexin_in_data
            # workingDirectorySuffix=UtilsImage.getPrefix(imageDict["image"][0]["path"]),
        )
        xds_indexing_task.execute()
        resultIndexing = None
        xparm_path = None
        spot_path = None
        if xds_indexing_task.isSuccess():
            xdsIndexingOutData = xds_indexing_task.outData
            if os.path.exists(xdsIndexingOutData["xparmXdsPath"]):
                xparm_path = xdsIndexingOutData["xparmXdsPath"]
            if os.path.exists(xdsIndexingOutData["spotXdsPath"]):
                spot_path = xdsIndexingOutData["spotXdsPath"]
            resultIndexing = ControlIndexing.getResultIndexingFromXds(
                xdsIndexingOutData
            )
        outData = {
            "resultIndexing": resultIndexing,
            "xparmXdsPath": xparm_path,
            "spotXdsPath": spot_path
        }
        return outData

    @staticmethod
    def getListSubWedge(in_data):
        list_sub_wedge = None
        # First check if we have data collection ids or image list
        # if "dataCollectionId" in inData:
        #     # TODO: get list of data collections from ISPyB
        #         logger.warning("Not implemented!")
        # el
        if "imagePath" in in_data or "fastCharacterisation":
            # Read the header(s)
            sub_wedge_assembly = SubWedgeAssembly(
                inData=in_data
                # workingDirectorySuffix=UtilsImage.getPrefix(listImagePath[0]),
            )
            sub_wedge_assembly.execute()
            list_sub_wedge = sub_wedge_assembly.outData["subWedge"]
        else:
            raise RuntimeError("No dataCollectionId or imagePath in inData")
        return list_sub_wedge

    @staticmethod
    def readImageHeaders(in_data):
        # Read the header(s)
        sub_wedge_assembly = SubWedgeAssembly(
            inData=in_data
            # workingDirectorySuffix=UtilsImage.getPrefix(listImagePath[0]),
        )
        sub_wedge_assembly.execute()
        list_sub_wedge = sub_wedge_assembly.outData["subWedge"]
        return list_sub_wedge

    @staticmethod
    def runControlDozor(listSubWedge):
        listControlDozor = []
        listOutDataControlDozor = []
        for subWedge in listSubWedge:
            listSubWedgeImage = subWedge["image"]
            for image in listSubWedgeImage:
                # listImage.append(image['path'])
                inDataControlDozor = {"image": [image["path"]]}
                controlDozor = ControlDozor(
                    inData=inDataControlDozor,
                    workingDirectorySuffix=UtilsImage.getPrefixNumber(image["path"]),
                )
                listControlDozor.append(controlDozor)
                controlDozor.start()
        for controlDozor in listControlDozor:
            controlDozor.join()
            if controlDozor.isSuccess():
                listOutDataControlDozor.append(controlDozor.outData)
        return listOutDataControlDozor

    @staticmethod
    def getResultIndexingFromXds(xdsIndexingOutData):
        idxref = xdsIndexingOutData["idxref"]
        xparamDict = xdsIndexingOutData["xparm"]
        if "A" in xparamDict:
            # Calculate MOSFLM UB matrix
            A = np.array(xparamDict["A"])
            B = np.array(xparamDict["B"])
            C = np.array(xparamDict["C"])

            volum = np.cross(A, B).dot(C)
            Ar = np.cross(B, C) / volum
            Br = np.cross(C, A) / volum
            Cr = np.cross(A, B) / volum
            UBxds = np.array([Ar, Br, Cr]).transpose()

            BEAM = np.array(xparamDict["beam"])
            ROT = np.array(xparamDict["rot"])
            wavelength = 1 / np.linalg.norm(BEAM)

            xparamDict["cell_volum"] = volum
            xparamDict["wavelength"] = wavelength
            xparamDict["Ar"] = Ar.tolist()
            xparamDict["Br"] = Br.tolist()
            xparamDict["Cr"] = Cr.tolist()
            xparamDict["UB"] = UBxds.tolist()

            normROT = float(np.linalg.norm(ROT))
            CAMERA_z = np.true_divide(ROT, normROT)
            CAMERA_y = np.cross(CAMERA_z, BEAM)
            normCAMERA_y = float(np.linalg.norm(CAMERA_y))
            CAMERA_y = np.true_divide(CAMERA_y, normCAMERA_y)
            CAMERA_x = np.cross(CAMERA_y, CAMERA_z)
            CAMERA = np.transpose(np.array([CAMERA_x, CAMERA_y, CAMERA_z]))

            mosflmUB = CAMERA.dot(UBxds) * xparamDict["wavelength"]
            # mosflmUB = UBxds*xparamDict["wavelength"]
            # xparamDict["mosflmUB"] = mosflmUB.tolist()

            reciprocCell = XDSIndexing.reciprocal(xparamDict["cell"])
            B = XDSIndexing.BusingLevy(reciprocCell)
            mosflmU = np.dot(mosflmUB, np.linalg.inv(B)) / xparamDict["wavelength"]
            # xparamDict[

            resultIndexing = {
                "spaceGroupNumber": idxref["spaceGroupNumber"],
                "cell": {
                    "a": idxref["a"],
                    "b": idxref["b"],
                    "c": idxref["c"],
                    "alpha": idxref["alpha"],
                    "beta": idxref["beta"],
                    "gamma": idxref["gamma"],
                },
                "xBeam": idxref["xBeam"],
                "yBeam": idxref["yBeam"],
                "distance": idxref["distance"],
                "qualityOfFit": idxref["qualityOfFit"],
                "mosaicity": idxref["mosaicity"],
                "XDS_xparm": xparamDict,
                "mosflmB": mosflmU.tolist(),
                "mosflmU": mosflmU.tolist(),
            }
        else:
            resultIndexing = {}
        return resultIndexing



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the 'Software'), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ['O. Svensson']
__license__ = 'MIT'
__date__ = '21/04/2019'

# Corresponding EDNA code:
# https://github.com/olofsvensson/edna-mx
# mxPluginExec/plugins/EDPluginH5ToCBF-v1.1/plugins/EDPluginH5ToCBFv1_1.py

import pathlib

from edna2.tasks.AbstractTask import AbstractTask

from edna2.utils import UtilsImage
from edna2.utils import UtilsConfig
from edna2.utils import UtilsLogging

logger = UtilsLogging.getLogger()


class H5ToCBFTask(AbstractTask):

    def getInDataSchema(self):
        return {
            "type": "object",
            "required": ["hdf5File"],
            "properties": {
                "imageNumber": {"type": "integer"},
                "startImageNumber": {"type": "integer"},
                "hdf5ImageNumber": {"type": "integer"},
                "hdf5File": {"type": "string"},
                "forcedOutputDirectory": {"type": "string"}
            }
        }

    def getOutDataSchema(self):
        return {
            "type": "object",
            "properties": {
                "outputCBFFile": {"type": "string"}
            }
        }

    def run(self, inData):
        outData = {}
        hdf5File = pathlib.Path(inData['hdf5File'])
        directory = hdf5File.parent
        prefix = UtilsImage.getPrefix(hdf5File)
        if 'imageNumber'in inData:
            commandLine, cbfFile = self.generateCommandsWithImageNumber(
                inData, directory, prefix, hdf5File)
            outData['outputCBFFile'] = str(cbfFile)
        elif 'startImageNumber' in inData and 'endImageNumber' in inData:
            commandLine, template = self.generateCommandsWithImageRange(
                inData, directory, prefix, hdf5File)
            outData['outputCBFFileTemplate'] = template
        self.setLogFileName('h5ToCBF.log')
        self.runCommandLine('/cvmfs/sb.esrf.fr/bin/eiger2cbf ' + commandLine, ignoreErrors=True)
        return outData

    @classmethod
    def generateCommandsWithImageNumber(cls, inData, directory, prefix,
                                        hdf5File):
        """
        This method creates a list of commands for the converter
        """
        imageNumber = inData['imageNumber']
        if 'hdf5ImageNumber' in inData:
            hdf5ImageNumber = inData['hdf5ImageNumber']
        else:
            hdf5ImageNumber = imageNumber
        if 'master.h5' in str(hdf5File):
            masterFile = hdf5File
        else:
            if UtilsConfig.isEMBL():
                fileName = '{0}_master.h5'.format(prefix)
            else:
                fileName = '{0}_{1}_master.h5'.format(prefix, hdf5ImageNumber)
            masterFile = directory / fileName
        if 'forcedOutputImageNumber' in inData:
            cbfFileName = prefix + \
                "_%04d" % inData['forcedOutputImageNumber'] + ".cbf"
            imageNumberInHdf5File = imageNumber
        else:
            cbfFileName = prefix + "_%04d" % imageNumber + ".cbf"
            imageNumberInHdf5File = 1
        if not 'forcedOutputDirectory' in inData:
            cbfFile = directory / cbfFileName
        else:
            forcedOutputDirectory = \
                pathlib.Path(inData['forcedOutputDirectory'])
            if not forcedOutputDirectory.exists():
                forcedOutputDirectory.mkdir(parents=True, mode=0o755)
            cbfFile = forcedOutputDirectory / cbfFileName
        commandLine = "{0} {1} {2}".format(
            masterFile, imageNumberInHdf5File, cbfFile
        )
        return commandLine, cbfFile

    @classmethod
    def generateCommandsWithImageRange(cls, inData, directory, prefix, hdf5File):
        startImageNumber = inData['startImageNumber']
        endImageNumber = inData['endImageNumber']
        if 'hdf5ImageNumber' in inData:
            hdf5ImageNumber = inData['hdf5ImageNumber']
        else:
            hdf5ImageNumber = startImageNumber
        if 'master.h5' in str(hdf5File):
            masterFile = hdf5File
        else:
            fileName = prefix + "_{0}_master.h5".format(hdf5ImageNumber)
            masterFile = directory / fileName
        cbfFileNamePrefix = prefix + '_'
        if 'forcedOutputDirectory' in inData:
            forcedOutputDirectory = \
                pathlib.Path(inData['forcedOutputDirectory'])
            if not forcedOutputDirectory.exists():
                forcedOutputDirectory.mkdir(mode=0o755, parents=True)
            cbfFilePath = forcedOutputDirectory / cbfFileNamePrefix
        else:
            cbfFilePath = directory / cbfFileNamePrefix
        commandLine = "{0} {1}:{2} {3}".format(
            masterFile, startImageNumber, endImageNumber, cbfFilePath)
        cbfFileTemplate = str(cbfFilePath) + "######.cbf"
        return commandLine, cbfFileTemplate





#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "21/04/2019"

# Corresponding EDNA code:
# https://github.com/olofsvensson/edna-mx
# mxv1/plugins/EDPluginGroupReadImageHeader-v1.0/plugins/
#      EDPluginControlReadImageHeaderv10.py

import os
import h5py
import numpy
import time
import pathlib

from edna2.utils import UtilsLogging

from edna2.tasks.AbstractTask import AbstractTask

from edna2.utils import UtilsPath
from edna2.utils import UtilsImage

logger = UtilsLogging.getLogger()

# Constants

# Default time out for wait file
DEFAULT_TIME_OUT = 30  # s

# Map between image suffix and image type
SUFFIX_ADSC = "img"
SUFFIX_MARCCD1 = "mccd"
SUFFIX_MARCCD2 = "marccd"
SUFFIX_Pilatus2M = "cbf"
SUFFIX_Pilatus6M = "cbf"
SUFFIX_Eiger4M = "cbf"
SUFFIX_Eiger9M = "cbf"
SUFFIX_Eiger16M = "cbf"

DICT_SUFFIX_TO_IMAGE = {
    SUFFIX_ADSC: "ADSC",
    SUFFIX_MARCCD1: "MARCCD",
    SUFFIX_MARCCD2: "MARCCD",
    SUFFIX_Pilatus2M: "Pilatus2M",
    SUFFIX_Pilatus6M: "Pilatus6M",
    SUFFIX_Eiger4M: "Eiger4M",
    SUFFIX_Eiger9M: "Eiger9M",
    SUFFIX_Eiger16M: "Eiger16M",
}


class ReadImageHeader(AbstractTask):

    def run(self, inData):
        listImagePath = inData["imagePath"]
        isFastMesh = inData.get("isFastMesh", False)
        hasOverlap = inData.get("hasOverlap", False)
        listSubWedge = []
        for imagePath in listImagePath:
            imageSuffix = os.path.splitext(imagePath)[1][1:]
            if imageSuffix == "cbf":
                subWedge = self.createCBFHeaderData(imagePath)
            elif imageSuffix == "h5":
                skipNumberOfImages = inData.get("skipNumberOfImages", False)
                subWedge = self.createHdf5HeaderData(
                    imagePath,
                    skipNumberOfImages,
                    isFastMesh=isFastMesh,
                    hasOverlap=hasOverlap,
                )
            else:
                raise RuntimeError(
                    "{0} cannot read image header from images with extension {1}".format(
                        self.__class__.__name__, imageSuffix
                    )
                )
            listSubWedge.append(subWedge)
        outData = {"subWedge": listSubWedge}
        return outData

    @classmethod
    def readCBFHeader(cls, filePath):
        """
        Returns an dictionary with the contents of a CBF image header.
        """
        dictHeader = None
        with open(filePath, "rb") as f:
            logger.info("Reading header from image " + filePath)
            f.seek(0, 0)
            doContinue = True
            iMax = 60
            index = 0
            while doContinue:
                line = f.readline().decode("utf-8")
                index += 1
                if "_array_data.header_contents" in line:
                    dictHeader = {}
                if "_array_data.data" in line or index > iMax:
                    doContinue = False
                if dictHeader is not None and line[0] == "#":
                    # Check for date
                    strTmp = line[2:].replace("\r\n", "")
                    if line[6] == "/" and line[10] == "/":
                        dictHeader["DateTime"] = strTmp
                    else:
                        strKey = strTmp.split(" ")[0]
                        strValue = strTmp.replace(strKey, "")[1:]
                        dictHeader[strKey] = strValue
        return dictHeader

    @classmethod
    def createCBFHeaderData(cls, imagePath):
        # Waiting for file
        timedOut, finalSize = UtilsPath.waitForFile(
            imagePath, expectedSize=100000, timeOut=DEFAULT_TIME_OUT
        )
        if timedOut:
            errorMessage = "Timeout when waiting for image %s" % imagePath
            logger.error(errorMessage)
            raise BaseException(errorMessage)
        dictHeader = cls.readCBFHeader(imagePath)
        detector = dictHeader["Detector:"]
        if (
            "PILATUS 3M" in detector
            or "PILATUS3 2M" in detector
            or "PILATUS 2M" in detector
            or "PILATUS2 3M" in detector
        ):
            detectorName = "PILATUS2 3M"
            detectorType = "pilatus2m"
            numberPixelX = 1475
            numberPixelY = 1679
        elif "PILATUS 6M" in detector or "PILATUS3 6M" in detector:
            detectorName = "PILATUS2 6M"
            detectorType = "pilatus6m"
            numberPixelX = 2463
            numberPixelY = 2527
        elif "eiger" in detector.lower() and "4m" in detector.lower():
            detectorName = "EIGER 4M"
            detectorType = "eiger4m"
            numberPixelX = 2070
            numberPixelY = 2167
        elif "eiger2" in detector.lower() and "9m" in detector.lower():
            detectorName = "EIGER2 9M"
            detectorType = "eiger9m"
            numberPixelX = 3108
            numberPixelY = 3262
        elif "eiger2" in detector.lower() and "16m" in detector.lower():
            detectorName = "EIGER2 16M"
            detectorType = "eiger16m"
            numberPixelX = 4148
            numberPixelY = 4362
        else:
            raise RuntimeError(
                "{0} cannot read image header from images with dector type {1}".format(
                    cls.__class__.__name__, detector
                )
            )
        experimentalCondition = {}
        detector = {"numberPixelX": numberPixelX, "numberPixelY": numberPixelY}
        # Pixel size
        listPixelSizeXY = dictHeader["Pixel_size"].split(" ")
        detector["pixelSizeX"] = float(listPixelSizeXY[0]) * 1000
        detector["pixelSizeY"] = float(listPixelSizeXY[3]) * 1000
        # Beam position
        listBeamPosition = (
            dictHeader["Beam_xy"]
            .replace("(", " ")
            .replace(")", " ")
            .replace(",", " ")
            .split()
        )
        detector["beamPositionX"] = float(listBeamPosition[0]) * detector["pixelSizeX"]
        detector["beamPositionY"] = float(listBeamPosition[1]) * detector["pixelSizeY"]
        distance = float(dictHeader["Detector_distance"].split(" ")[0]) * 1000
        detector["distance"] = distance
        detector["serialNumber"] = dictHeader["Detector:"]
        detector["name"] = detectorName
        detector["type"] = detectorType
        experimentalCondition["detector"] = detector
        # Beam object
        beam = {
            "wavelength": float(dictHeader["Wavelength"].split(" ")[0]),
            "exposureTime": float(dictHeader["Exposure_time"].split(" ")[0]),
        }
        experimentalCondition["beam"] = beam
        # Goniostat object
        goniostat = {}
        rotationAxisStart = float(dictHeader["Start_angle"].split(" ")[0])
        oscillationWidth = float(dictHeader["Angle_increment"].split(" ")[0])
        goniostat["rotationAxisStart"] = rotationAxisStart
        goniostat["rotationAxisEnd"] = rotationAxisStart + oscillationWidth
        goniostat["oscillationWidth"] = oscillationWidth
        experimentalCondition["goniostat"] = goniostat
        # Create the image object
        image = {"path": imagePath}
        if "DateTime" in dictHeader:
            image["date"] = dictHeader["DateTime"]
        imageNumber = UtilsImage.getImageNumber(imagePath)
        image["number"] = imageNumber
        subWedge = {"experimentalCondition": experimentalCondition, "image": [image]}
        return subWedge

    @classmethod
    def readHdf5Header(cls, filePath):
        """
        Returns an dictionary with the contents of an Eiger Hdf5 image header.
        """
        logger.info("Reading header from image " + str(filePath))
        f = h5py.File(filePath, "r")
        dictHeader = {
            "wavelength": f["entry"]["instrument"]["beam"]["incident_wavelength"][()],
            "beam_center_x": f["entry"]["instrument"]["detector"]["beam_center_x"][()],
            "beam_center_y": f["entry"]["instrument"]["detector"]["beam_center_y"][()],
            "count_time": f["entry"]["instrument"]["detector"]["count_time"][()],
            "detector_distance": f["entry"]["instrument"]["detector"][
                "detector_distance"
            ][()],
            "translation": list(
                f["entry"]["instrument"]["detector"]["geometry"]["translation"][
                    "distances"
                ]
            ),
            "x_pixel_size": f["entry"]["instrument"]["detector"]["x_pixel_size"][()],
            "y_pixel_size": f["entry"]["instrument"]["detector"]["y_pixel_size"][()],
            "omega_range_average": f["entry"]["sample"]["goniometer"][
                "omega_range_average"
            ][()],
            "detector_number": f["entry"]["instrument"]["detector"]["detector_number"][
                ()
            ].decode("utf-8"),
            "description": f["entry"]["instrument"]["detector"]["description"][
                ()
            ].decode("utf-8"),
            "data_collection_date": f["entry"]["instrument"]["detector"][
                "detectorSpecific"
            ]["data_collection_date"][()].decode("utf-8"),
            "data": list(f["entry"]["data"]),
        }
        # 'Old' Eiger files have just one entry for 'omega'
        omega = f["entry"]["sample"]["goniometer"]["omega"][()]
        if type(omega) == numpy.float32:
            dictHeader["omega_start"] = float(omega)
        else:
            dictHeader["omega_start"] = float(omega[0])
        f.close()
        return dictHeader

    @classmethod
    def createHdf5HeaderData(
        cls, imagePath, skipNumberOfImages=False, hasOverlap=False, isFastMesh=True
    ):
        h5MasterFilePath, h5DataFilePath, h5FileNumber = UtilsImage.getH5FilePath(
            pathlib.Path(imagePath), isFastMesh=isFastMesh, hasOverlap=hasOverlap
        )
        # Waiting for file
        timedOut, finalSize = UtilsPath.waitForFile(
            h5MasterFilePath, expectedSize=2000000, timeOut=DEFAULT_TIME_OUT
        )
        if timedOut:
            errorMessage = "Timeout when waiting for image %s" % imagePath
            logger.error(errorMessage)
            raise BaseException(errorMessage)
        logger.info("Final size for {0}: {1}".format(h5MasterFilePath, finalSize))
        noTrialsLeft = 5
        dictHeader = None
        while noTrialsLeft > 0:
            try:
                dictHeader = cls.readHdf5Header(h5MasterFilePath)
                noTrialsLeft = 0
            except Exception as e:
                logger.warning(
                    "Cannot read header from {0}, no trials left: {1}".format(
                        h5MasterFilePath, noTrialsLeft
                    )
                )
                time.sleep(5)
                noTrialsLeft -= 1
        if dictHeader is None:
            raise RuntimeError("Cannot read header from {0}!".format(h5MasterFilePath))
        description = dictHeader["description"]
        if "Eiger 4M" in description:
            detectorName = "EIGER 4M"
            detectorType = "eiger4m"
            numberPixelX = 2070
            numberPixelY = 2167
        elif "eiger2" in description.lower() and "9m" in description.lower():
            detectorName = "EIGER2 9M"
            detectorType = "eiger9m"
            numberPixelX = 3108
            numberPixelY = 3262
        elif "eiger" in description.lower() and "16M" in description:
            detectorName = "EIGER 16M"
            detectorType = "eiger16m"
            numberPixelX = 4148
            numberPixelY = 4362
        else:
            raise RuntimeError(
                "{0} cannot read image header from images with detector type {1}".format(
                    cls.__class__.__name__, description
                )
            )
        # Image number
        image_number = UtilsImage.getImageNumber(imagePath)
        # Find out size of data set
        prefix = str(h5MasterFilePath).split("master.h5")[0]
        listDataImage = []
        noImages = 0
        if not skipNumberOfImages:
            for data in dictHeader["data"]:
                dataFilePath = prefix + data + ".h5"
                timedOut, finalSize = UtilsPath.waitForFile(
                    dataFilePath, expectedSize=100000, timeOut=DEFAULT_TIME_OUT
                )
                if timedOut:
                    raise RuntimeError(
                        "Timeout waiting for file {0}".format(dataFilePath)
                    )
                # listDataImage.append({
                #     'path': dataFilePath
                # })
                f = h5py.File(dataFilePath, "r")
                dataShape = f["entry"]["data"]["data"].shape
                noImages += dataShape[0]
                f.close()
        experimentalCondition = {}
        # Pixel size and beam position
        detector = {
            "numberPixelX": int(numberPixelX),
            "numberPixelY": int(numberPixelY),
            "pixelSizeX": round(dictHeader["x_pixel_size"] * 1000, 3),
            "pixelSizeY": round(dictHeader["y_pixel_size"] * 1000, 3),
            "beamPositionX": round(
                float(dictHeader["beam_center_x"] * dictHeader["x_pixel_size"] * 1000),
                3,
            ),
            "beamPositionY": round(
                float(dictHeader["beam_center_y"] * dictHeader["y_pixel_size"] * 1000),
                3,
            ),
            "distance": round(float(dictHeader["detector_distance"]) * 1000, 3),
            "serialNumber": dictHeader["detector_number"],
            "name": detectorName,
            "type": detectorType,
        }
        experimentalCondition["detector"] = detector
        # Beam object
        beam = {
            "wavelength": round(float(dictHeader["wavelength"]), 6),
            "exposureTime": round(float(dictHeader["count_time"]), 6),
        }
        experimentalCondition["beam"] = beam
        # Goniostat object
        goniostat = {}
        rotationAxisStart = round(float(dictHeader["omega_start"]), 4)
        oscillationWidth = round(float(dictHeader["omega_range_average"]), 4)
        # Offset for the image number
        rotationAxisStart += (image_number - 1)*oscillationWidth
        goniostat["rotationAxisStart"] = rotationAxisStart
        goniostat["rotationAxisEnd"] = rotationAxisStart + oscillationWidth
        goniostat["oscillationWidth"] = oscillationWidth
        experimentalCondition["goniostat"] = goniostat
        # Create the image object
        image_dict = {
            "path": imagePath,
            "date": dictHeader["data_collection_date"],
            "number": 1,
        }
        # imageNumber = UtilsImage.getImageNumber(imagePath)
        # image['number'] = imageNumber
        subWedge = {
            "experimentalCondition": experimentalCondition,
            "image": [image_dict],
        }
        return subWedge



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from __future__ import division, print_function
import os
import sys
import json
import jsonschema
import pathlib
import numpy as np
import scipy.spatial as sp

from cctbx import miller
from cctbx import crystal

from edna2.tasks.AbstractTask import AbstractTask
from edna2.utils import UtilsLogging

__author__ = ['S. Basu']
__license__ = 'MIT'
__date__ = '20/11/2020'

"Inspired from Computing workshop in ECM30 conference: https://github.com/keitaroyam/ECACOMSIG-ECM30"

logger = UtilsLogging.getLogger()


class ExecSpotPrediction(AbstractTask):

    def __init__(self, jData):
        AbstractTask.__init__(self, inData=jData)
        self.xparamdict = self.digestxparamFile(inData=jData)

    def getInDataSchema(self):
        return {
            "image": {"type": "string"},
            "XDSspotFile": {"type", "string"},
            "XDSidxrefFile": {"type", "string"},
            "XDSxparamFile": {"type", "string"},
            "resolution": {"type", "string"},
            "frame": {"type", "integer"},
            "mosaicity": {"type", "number"}
        }

    @staticmethod
    def nearest_neighbor(spot_array, predicted_array, dist_limit, ofh):
        kdtree = sp.KDTree(spot_array)
        nNeighbor = 0
        for (xcal, ycal, phi, zeta) in predicted_array:
            d, idx = kdtree.query((xcal, ycal), k=1, p=2)
            print(d, idx)
            if idx == kdtree.n:
                continue
            if d < dist_limit:
                ofh.write("%.2f %.2f \n" % tuple(spot_array[idx]))
                nNeighbor += 1

        return nNeighbor

    @staticmethod
    def digestxparamFile(inData):
        xparaminfo = dict()
        idxfile = pathlib.Path(inData['XDSxparamFile'])
        if not idxfile.exists():
            return xparaminfo
        idx = open(idxfile, 'r')
        allLines = idx.readlines()
        for ii in range(len(allLines)):
            if ii == 1:
                line = allLines[ii].split()
                xparaminfo['frame_no'] = int(line[0])
                xparaminfo['start_angle'] = float(line[1])
                xparaminfo['osc_range'] = float(line[2])
                xparaminfo['gonio_coord'] = list(map(float, line[3:6]))
            elif ii == 2:
                line = allLines[ii].split()
                xparaminfo['wavelength'] = float(line[0])
                xparaminfo['incident_beam'] = list(map(float, line[1:4]))
            elif ii == 3:
                line = allLines[ii].split()
                xparaminfo['spacegroup'] = str(line[0])
                xparaminfo['unit_cell'] = list(map(float, line[1:7]))
            elif ii == 4:
                line = allLines[ii].split()
                xparaminfo['a_axis'] = list(map(float, line))
            elif ii == 5:
                line = allLines[ii].split()
                xparaminfo['b_axis'] = list(map(float, line))
            elif ii == 6:
                line = allLines[ii].split()
                xparaminfo['c_axis'] = list(map(float, line))
            elif ii == 7:
                line = allLines[ii].split()
                xparaminfo['NX'] = int(line[1])
                xparaminfo['NY'] = int(line[2])
                xparaminfo['pixelsize'] = float(line[3])
            elif ii == 8:
                line = allLines[ii].split()
                xparaminfo['beamX'] = float(line[0])
                xparaminfo['beamY'] = float(line[1])
                xparaminfo['detZ'] = float(line[2])
            elif ii == 9:
                line = allLines[ii].split()
                xparaminfo['det_axis1'] = list(map(float, line))
            elif ii == 10:
                xparaminfo['det_axis2'] = list(map(float, allLines[ii].split()))
            elif ii == 11:
                xparaminfo['det_axis3'] = list(map(float, allLines[ii].split()))
            else:
                pass
        return xparaminfo

    def collect_spots(self):
        spotfile = pathlib.Path(self.inData.get('XDSspotFile', ''))
        spot_list = []
        if not spotfile.exists():
            return spot_list
        with open(spotfile, 'r') as fh:
            for line in fh:
                x, y = map(lambda s: float(s), line.strip().split()[:2])
                spot_list.append((x, y))
        return spot_list

    def predict_miller_indices(self):
        crystal_symmetry = crystal.symmetry(unit_cell=self.xparamdict['unit_cell'],
                                            space_group_symbol=self.xparamdict['spacegroup'])
        d_min = self.inData.get('resolution', float('3'))
        d_max = None
        miller_set = miller.build_set(crystal_symmetry, anomalous_flag=True, d_min=d_min, d_max=d_max)
        return miller_set.indices()

    def calc_detector_projection(self):
        h_mat = self.predict_miller_indices()  # get predicted set of miller indices
        h_mat = np.array(h_mat)
        cell_mat = np.array([self.xparamdict['a_axis'], self.xparamdict['b_axis'], self.xparamdict['c_axis']]).transpose()
        rlp = np.linalg.inv(cell_mat)  # (a*, b*, c*) matrix

        # incident beam vector
        s0 = np.array(self.xparamdict['incident_beam'])
        beam_vector = s0 / np.linalg.norm(s0) / self.xparamdict['wavelength']

        rot_vector = np.array(self.xparamdict['gonio_coord'])  # obtained from rotation axis
        rot = np.empty((3, 3), dtype=np.float)
        rot[:, 1] = rot_vector / np.linalg.norm(rot_vector)
        rot[:, 0] = np.cross(rot[:, 1], beam_vector)
        rot[:, 0] /= np.linalg.norm(rot[:, 0])
        rot[:, 2] = np.cross(rot[:, 0], rot[:, 1])
        # detector coordinate system
        d_mat = np.array([self.xparamdict['det_axis1'], self.xparamdict['det_axis2'], self.xparamdict['det_axis3']])
        d_mat = d_mat.transpose()

        # 1st predicted spot as vector on Ewald sphere
        p0star = np.dot(h_mat, rlp)
        p0star_gonio = np.dot(p0star, rot)  # same predicted spot on gonio coordinate
        beam_gonio = np.dot(rot.transpose(), beam_vector)  # incident beam on gonio coordinate
        print(beam_gonio.shape)
        p0star_sq_dist = np.sum(p0star**2, axis=1)  # sq distance of predicted spot from the rotation axis

        pstar_gonio = np.empty(p0star_gonio.shape)  # a 3x3 matrix
        pstar_gonio[:, 2] = (-0.5*p0star_sq_dist - beam_gonio[1]*p0star_gonio[:, 1]) / beam_gonio[2]
        pstar_gonio[:, 1] = p0star_gonio[:, 1]
        pstar_gonio[:, 0] = p0star_sq_dist - pstar_gonio[:, 1]**2 - pstar_gonio[:, 2]**2  # sqrt taken at later stage

        sel_index = pstar_gonio[:, 0] > 0  # exclude blind region with this criteria, filters the matrices

        h_mat = h_mat[sel_index]
        p0star_gonio = p0star_gonio[sel_index]
        pstar_gonio = pstar_gonio[sel_index]

        pstar_gonio[:, 0] = np.sqrt(pstar_gonio[:, 0])

        projected_hkl = np.empty((0, 3), dtype=np.int)  # h,k,l as row vector
        position_on_detector = np.empty((0, 4), dtype=np.float)  # Find detector projection, x, y, phi, zeta

        for sign in (+1, -1):
            pstar_gonio[:, 0] *= sign
            phi = np.arctan2((p0star_gonio[:, 2]*pstar_gonio[:, 0] - p0star_gonio[:, 1]*pstar_gonio[:, 2]),
                             (p0star_gonio[:, 1]*pstar_gonio[:, 1] + p0star_gonio[:, 2]*pstar_gonio[:, 2]))

            refl_std = np.cross(pstar_gonio, beam_gonio)
            refl_std /= np.linalg.norm(refl_std, axis=1).reshape(refl_std.shape[0], 1)  # convert into column vector
            zeta = refl_std[:, 1]

            Svector = beam_vector + np.dot(pstar_gonio, rot.transpose())
            Svec_on_det = np.dot(Svector, d_mat)
            sel_index = self.xparamdict['detZ']*Svec_on_det[:, 2] > 0
            Svec_on_det, h_mat, phi, zeta = Svec_on_det[sel_index], h_mat[sel_index], phi[sel_index], zeta[sel_index]

            det_pos_x = self.xparamdict['beamX'] + self.xparamdict['detZ']*Svec_on_det[:, 0] / Svec_on_det[:, 2] \
                        / self.xparamdict['pixelsize']

            det_pos_y = self.xparamdict['beamY'] + self.xparamdict['detZ'] * Svec_on_det[:, 1] / Svec_on_det[:, 2] \
                        / self.xparamdict['pixelsize']

            projected_hkl = np.row_stack([projected_hkl, h_mat])
            position_on_detector = np.row_stack([position_on_detector,
                                                 np.column_stack([det_pos_x, det_pos_y, phi, zeta])])

        return projected_hkl, position_on_detector

    def get_spot_prediction(self, esd=3.0):

        mosaicity = self.inData.get('mosaicity', float('0.1'))
        frame = self.inData.get('frame', int(1))
        phi_from_expt = self.xparamdict['start_angle'] + \
                            self.xparamdict['osc_range']*(frame - self.xparamdict['start_angle'] + 0.5)
        phi_from_expt, mosaicity, osc_range = np.deg2rad([phi_from_expt, mosaicity, self.xparamdict['osc_range']])

        hkl_calc, projection = self.calc_detector_projection()
        phi_diff = np.fmod(projection[:, 2] - phi_from_expt, 2.0*np.pi)
        # put all phi differences within -pi to +pi range..

        phi_diff[phi_diff < - np.pi] += 2.0 * np.pi
        phi_diff[phi_diff > np.pi] -= 2.0 * np.pi
        # choose range of index to visualize..
        sel_index = np.abs(phi_diff) < (osc_range / 2. + esd * mosaicity / np.abs(projection[:, 3]))

        return hkl_calc[sel_index], projection[sel_index]


if __name__ == '__main__':
    inData = {
        "XDSspotFile": "/Users/shbasu/work/APICA/1x/SPOT.XDS",
        "XDSxparamFile": "/Users/shbasu/work/APICA/1x/XPARM.XDS"
    }

    pp = ExecSpotPrediction(inData)
    # xparam = pp.digestxparamFile(inData=inData)
    pmiller, pdata = pp.get_spot_prediction()
    spots = pp.collect_spots()
    distance_limit_in_pixel = 20

    out = open("tr1.adx", "w")
    nmatches = pp.nearest_neighbor(np.array(spots), pdata, distance_limit_in_pixel, out)
    '''
    with open("tst.adx", 'w') as ofh:
        for (h, k, l), (x, y, phi, zeta) in zip(pmiller, pdata):
            ofh.write('%d %d %d %d %d\n' % (x, y, h, k, l))
    '''

